# SPDX-FileCopyrightText: 2022-present Inria
# SPDX-FileCopyrightText: 2022-present Alexandre Abadie <alexandre.abadie@inria.fr>
# SPDX-FileCopyrightText: 2023-present Filip Maksimovic <filip.maksimovic@inria.fr>
# SPDX-FileCopyrightText: 2024-present Diego Badillo <diego.badillo@sansano.usm.cl>
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module for the Dotbot protocol API."""

import dataclasses
import typing
from abc import ABC
from binascii import hexlify
from dataclasses import dataclass
from enum import IntEnum
from typing import List

PROTOCOL_VERSION = 1
PAYLOAD_RESERVED_THRESHOLD = 0x80


class PayloadType(IntEnum):
    """Types of DotBot payload types."""

    CMD_MOVE_RAW = 0x00
    CMD_RGB_LED = 0x01
    LH2_RAW_LOCATION = 0x02
    LH2_LOCATION = 0x03
    ADVERTISEMENT = 0x04
    GPS_POSITION = 0x05
    DOTBOT_DATA = 0x06
    CONTROL_MODE = 0x07
    LH2_WAYPOINTS = 0x08
    GPS_WAYPOINTS = 0x09
    SAILBOT_DATA = 0x0A
    CMD_XGO_ACTION = 0x0B
    LH2_PROCESSED_DATA = 0x0C
    LH2_RAW_DATA = 0x0D
    RAW_DATA = 0x10
    DOTBOT_SIMULATOR_DATA = 0xFA


class ApplicationType(IntEnum):
    """Types of DotBot applications."""

    DotBot = 0  # pylint: disable=invalid-name
    SailBot = 1  # pylint: disable=invalid-name
    Freebot = 2  # pylint: disable=invalid-name
    XGO = 3
    LH2_mini_mote = 4


class ControlModeType(IntEnum):
    """Types of DotBot control modes."""

    MANUAL = 0
    AUTO = 1


class ProtocolPayloadParserException(Exception):
    """Exception raised on invalid or unsupported payload."""


class PacketType(IntEnum):
    """Types of MAC layer packet."""

    BEACON = 1
    JOIN_REQUEST = 2
    JOIN_RESPONSE = 4
    KEEP_ALIVE = 8
    DATA = 16


@dataclass
class PacketFieldMetadata:
    """Data class that describes a packet field metadata."""

    name: str = ""
    disp: str = ""
    length: int = 1
    signed: bool = False
    type_: typing.Any = int

    def __post_init__(self):
        if not self.disp:
            self.disp = self.name


@dataclass
class Packet(ABC):
    """Base class for packet classes."""

    @property
    def size(self) -> int:
        return sum(field.length for field in self.metadata)

    def from_bytes(self, bytes_):
        fields = dataclasses.fields(self)
        # base class makes metadata attribute mandatory so there's at least one
        # field defined in subclasses
        # first elements in fields has to be metadata
        if not fields or fields[0].name != "metadata":
            raise ValueError("metadata must be defined first")
        metadata = fields[0].default_factory()
        for idx, field in enumerate(fields[1:]):
            if metadata[idx].type_ is list:
                element_class = typing.get_args(field.type)[0]
                field_attribute = getattr(self, field.name)
                # subclass element is a list and previous attribute is called
                # "count" and should have already been retrieved from the byte
                # stream
                for _ in range(self.count):
                    element = element_class()
                    if len(bytes_) < element.size:
                        raise ValueError("Not enough bytes to parse")
                    field_attribute.append(element.from_bytes(bytes_))
                    bytes_ = bytes_[element.size :]
            elif metadata[idx].type_ in [bytes, bytearray]:
                # subclass element is bytes and previous attribute is called
                # "count" and should have already been retrieved from the byte
                # stream
                length = metadata[idx].length
                if hasattr(self, "count"):
                    length = self.count
                setattr(self, field.name, bytes_[0:length])
                bytes_ = bytes_[length:]
            else:
                length = metadata[idx].length
                if len(bytes_) < length:
                    raise ValueError("Not enough bytes to parse")
                setattr(
                    self,
                    field.name,
                    int.from_bytes(
                        bytes=bytes_[0:length],
                        signed=metadata[idx].signed,
                        byteorder="little",
                    ),
                )
                bytes_ = bytes_[length:]
        return self

    def to_bytes(self, byteorder="little") -> bytes:
        buffer = bytearray()
        metadata = dataclasses.fields(self)[0].default_factory()
        for idx, field in enumerate(dataclasses.fields(self)[1:]):
            value = getattr(self, field.name)
            if isinstance(value, list):
                for element in value:
                    buffer += element.to_bytes()
            elif isinstance(value, (bytes, bytearray)):
                buffer += value
            else:
                buffer += int(value).to_bytes(
                    length=metadata[idx].length,
                    byteorder=byteorder,
                    signed=metadata[idx].signed,
                )
        return buffer


@dataclass
class Header(Packet):
    """Dataclass that holds MAC header fields."""

    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [
            PacketFieldMetadata(name="version", disp="ver.", length=1),
            PacketFieldMetadata(name="type_", disp="type", length=1),
            PacketFieldMetadata(name="destination", disp="dst", length=8),
            PacketFieldMetadata(name="source", disp="src", length=8),
        ]
    )
    version: int = PROTOCOL_VERSION
    type_: int = PacketType.DATA
    destination: int = 0xFFFFFFFFFFFFFFFF
    source: int = 0x0000000000000000


@dataclass
class PayloadAdvertisement(Packet):
    """Dataclass that holds an advertisement (emtpy)."""

    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [
            PacketFieldMetadata(name="application", disp="app"),
        ]
    )

    application: ApplicationType = ApplicationType.DotBot


@dataclass
class PayloadCommandMoveRaw(Packet):
    """Dataclass that holds move raw command data fields."""

    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [
            PacketFieldMetadata(name="lelf_x", disp="lx", signed=True),
            PacketFieldMetadata(name="lelf_y", disp="ly", signed=True),
            PacketFieldMetadata(name="right_y", disp="rx", signed=True),
            PacketFieldMetadata(name="right_y", disp="ry", signed=True),
        ]
    )

    left_x: int = 0
    left_y: int = 0
    right_x: int = 0
    right_y: int = 0


@dataclass
class PayloadCommandRgbLed(Packet):
    """Dataclass that holds a complete rgb led command fields."""

    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [
            PacketFieldMetadata(name="red"),
            PacketFieldMetadata(name="green"),
            PacketFieldMetadata(name="blue"),
        ]
    )

    red: int = 0
    green: int = 0
    blue: int = 0


@dataclass
class PayloadCommandXgoAction(Packet):
    """Dataclass that holds an XGO action."""

    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [
            PacketFieldMetadata(name="action"),
        ]
    )
    action: int = 0


@dataclass
class PayloadLh2RawLocation(Packet):
    """Dataclass that holds LH2 raw location data."""

    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [
            PacketFieldMetadata(name="bits", length=8),
            PacketFieldMetadata(name="polynomial_index", disp="poly", length=1),
            PacketFieldMetadata(name="offset", disp="off.", length=1, signed=True),
        ]
    )

    bits: int = 0x0000000000000000
    polynomial_index: int = 0x00
    offset: int = 0x00


@dataclass
class PayloadLh2RawData(Packet):
    """Dataclass that holds LH2 raw data."""

    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [
            PacketFieldMetadata(name="count", disp="len"),
            PacketFieldMetadata(name="locations", type_=list, length=0),
        ]
    )

    count: int = 0
    locations: list[PayloadLh2RawLocation] = dataclasses.field(
        default_factory=lambda: []
    )


@dataclass
class PayloadLh2ProcessedLocation(Packet):
    """Dataclass that holds LH2 processed location data."""

    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [
            PacketFieldMetadata(name="polynomial_index", disp="poly"),
            PacketFieldMetadata(name="lfsr_index", length=4),
            PacketFieldMetadata(name="timestamp_us", length=4),
        ]
    )

    polynomial_index: int = 0x00
    lfsr_index: int = 0x00000000
    timestamp_us: int = 0x00000000


@dataclass
class PayloadLH2Location(Packet):
    """Dataclass that holds LH2 computed location data."""

    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [
            PacketFieldMetadata(name="pos_x", disp="x", length=4),
            PacketFieldMetadata(name="pos_y", disp="y", length=4),
            PacketFieldMetadata(name="pos_z", disp="z", length=4),
        ]
    )

    pos_x: int = 0
    pos_y: int = 0
    pos_z: int = 0


@dataclass
class PayloadDotBotData(Packet):
    """Dataclass that holds direction and LH2 raw data from DotBot application."""

    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [
            PacketFieldMetadata(name="direction", disp="dir.", length=2, signed=True),
            PacketFieldMetadata(name="count", disp="len"),
            PacketFieldMetadata(name="locations", type_=list, length=0),
        ]
    )

    direction: int = 0xFFFF
    count: int = 0
    locations: List[PayloadLh2RawLocation] = dataclasses.field(
        default_factory=lambda: []
    )


@dataclass
class PayloadGPSPosition(Packet):
    """Dataclass that holds GPS positions."""

    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [
            PacketFieldMetadata(name="latitude", disp="lat.", length=4, signed=True),
            PacketFieldMetadata(name="longitude", disp="long.", length=4, signed=True),
        ]
    )

    latitude: int = 0
    longitude: int = 0


@dataclass
class PayloadSailBotData(Packet):
    """Dataclass that holds SailBot data from SailBot application."""

    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [
            PacketFieldMetadata(name="direction", disp="dir.", length=2, signed=False),
            PacketFieldMetadata(name="latitude", disp="lat.", length=4, signed=True),
            PacketFieldMetadata(name="longitude", disp="long.", length=4, signed=True),
            PacketFieldMetadata(name="wind_angle", disp="wind", length=2, signed=False),
            PacketFieldMetadata(name="rudder_angle", disp="rud.", signed=True),
            PacketFieldMetadata(name="sail_angle", disp="sail.", signed=True),
        ]
    )

    direction: int = 0xFFFF
    latitude: int = 0
    longitude: int = 0
    wind_angle: int = 0xFFFF
    rudder_angle: int = 0
    sail_angle: int = 0


@dataclass
class PayloadDotBotSimulatorData(Packet):
    """Dataclass that holds direction and GPS data and heading from SailBot application."""

    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [
            PacketFieldMetadata(name="theta", length=2),
            PacketFieldMetadata(name="pos_x", length=4),
            PacketFieldMetadata(name="pos_y", length=4),
        ]
    )

    theta: int = 0xFFFF
    pos_x: int = 0
    pos_y: int = 0


@dataclass
class PayloadControlMode(Packet):
    """Dataclass that holds a control mode message."""

    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [PacketFieldMetadata(name="mode", disp="mode")]
    )

    mode: ControlModeType = ControlModeType.MANUAL


@dataclass
class PayloadLH2Waypoints(Packet):
    """Dataclass that holds a list of LH2 waypoints."""

    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [
            PacketFieldMetadata(name="threshold", disp="thr."),
            PacketFieldMetadata(name="count", disp="len."),
            PacketFieldMetadata(name="waypoints", type_=list, length=0),
        ]
    )

    threshold: int = 0
    count: int = 0
    waypoints: list[PayloadLH2Location] = dataclasses.field(default_factory=lambda: [])


@dataclass
class PayloadGPSWaypoints(Packet):
    """Dataclass that holds a list of GPS waypoints."""

    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [
            PacketFieldMetadata(name="threshold", disp="thr."),
            PacketFieldMetadata(name="count", disp="len."),
            PacketFieldMetadata(name="waypoints", type_=list, length=0),
        ]
    )

    threshold: int = 0
    count: int = 0
    waypoints: list[PayloadGPSPosition] = dataclasses.field(default_factory=lambda: [])


@dataclass
class PayloadRawData(Packet):
    """Dataclass that holds raw bytes data."""

    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [
            PacketFieldMetadata(name="count", disp="len."),
            PacketFieldMetadata(name="data", type_=bytes, length=0),
        ]
    )

    count: int = 0
    data: bytes = dataclasses.field(default_factory=lambda: bytearray)


PAYLOAD_PARSERS: dict[int, Packet] = {
    PayloadType.ADVERTISEMENT: PayloadAdvertisement,
    PayloadType.CMD_MOVE_RAW: PayloadCommandMoveRaw,
    PayloadType.CMD_RGB_LED: PayloadCommandRgbLed,
    PayloadType.CMD_XGO_ACTION: PayloadCommandXgoAction,
    PayloadType.LH2_RAW_LOCATION: PayloadLh2RawLocation,
    PayloadType.LH2_PROCESSED_DATA: PayloadLh2ProcessedLocation,
    PayloadType.LH2_RAW_DATA: PayloadLh2RawData,
    PayloadType.LH2_LOCATION: PayloadLH2Location,
    PayloadType.DOTBOT_DATA: PayloadDotBotData,
    PayloadType.GPS_POSITION: PayloadGPSPosition,
    PayloadType.SAILBOT_DATA: PayloadSailBotData,
    PayloadType.DOTBOT_SIMULATOR_DATA: PayloadDotBotSimulatorData,
    PayloadType.CONTROL_MODE: PayloadControlMode,
    PayloadType.LH2_WAYPOINTS: PayloadLH2Waypoints,
    PayloadType.GPS_WAYPOINTS: PayloadGPSWaypoints,
    PayloadType.RAW_DATA: PayloadRawData,
}


def register_parser(payload_type: int, parser: Packet):
    """Register a new payload parser."""
    if payload_type in PAYLOAD_PARSERS:
        raise ValueError(f"Payload type '0x{payload_type:02X}' already registered")
    if payload_type < PAYLOAD_RESERVED_THRESHOLD:
        raise ValueError(f"Payload type '0x{payload_type:02X}' is reserved")
    PAYLOAD_PARSERS[payload_type] = parser


@dataclass
class Frame:
    """Data class that holds a payload packet."""

    header: Header = None
    payload: Packet = None

    @property
    def payload_type(self) -> int:
        for payload_type, cls_ in PAYLOAD_PARSERS.items():
            if cls_ == self.payload.__class__:
                return payload_type
        raise ValueError(f"Unsupported payload class '{self.payload.__class__}'")

    def from_bytes(self, bytes_):
        self.header = Header().from_bytes(bytes_[0:18])
        payload_type = int.from_bytes(bytes_[18:19], "little")
        if payload_type not in PAYLOAD_PARSERS:
            raise ProtocolPayloadParserException(
                f"Unsupported payload type '{payload_type}'"
            )
        self.payload = PAYLOAD_PARSERS[payload_type]().from_bytes(bytes_[19:])
        return self

    def to_bytes(self, byteorder="little") -> bytes:
        header_bytes = self.header.to_bytes(byteorder)
        payload_bytes = self.payload.to_bytes(byteorder)
        return header_bytes + int.to_bytes(self.payload_type) + payload_bytes

    def __repr__(self):
        header_separators = [
            "-" * (2 * field.length + 4) for field in self.header.metadata
        ]
        type_separators = ["-" * 6]
        payload_separators = [
            "-" * (2 * field.length + 4)
            for field in self.payload.metadata
            if field.type_ is int
        ]
        payload_separators += [
            "-" * (2 * field_metadata.length + 4)
            for metadata in self.payload.metadata
            if metadata.type_ is list
            for field in getattr(self.payload, metadata.name)
            for field_metadata in field.metadata
        ]
        payload_separators += [
            "-" * (2 * len(getattr(self.payload, field.name)) + 4)
            for field in self.payload.metadata
            if field.type_ is bytes
        ]
        header_names = [
            f" {field.disp:<{2 * field.length + 3}}" for field in self.header.metadata
        ]
        payload_names = [
            f" {field.disp:<{2 * field.length + 3}}"
            for field in self.payload.metadata
            if field.type_ in (int, bytes) and field.length > 0
        ]
        payload_names += [
            f" {field.disp:<{2 * len(getattr(self.payload, field.name)) + 3}}"
            for field in self.payload.metadata
            if field.type_ is bytes and field.length == 0
        ]
        payload_names += [
            f" {field_metadata.disp:<{2 * field_metadata.length + 3}}"
            for metadata in self.payload.metadata
            if metadata.type_ is list
            for field in getattr(self.payload, metadata.name)
            for field_metadata in field.metadata
        ]
        header_values = [
            f" 0x{hexlify(int(getattr(self.header, field.name)).to_bytes(self.header.metadata[idx].length, 'big', signed=self.header.metadata[idx].signed)).decode():<{2 * self.header.metadata[idx].length + 1}}"
            for idx, field in enumerate(dataclasses.fields(self.header)[1:])
        ]
        type_value = [f" 0x{hexlify(self.payload_type.to_bytes(1, 'big')).decode():<3}"]
        payload_values = [
            f" 0x{hexlify(int(getattr(self.payload, field.name)).to_bytes(self.payload.metadata[idx].length, 'big', signed=self.payload.metadata[idx].signed)).decode():<{2 * self.payload.metadata[idx].length + 1}}"
            for idx, field in enumerate(dataclasses.fields(self.payload)[1:])
            if self.payload.metadata[idx].type_ is int
        ]
        payload_values += [
            f" 0x{hexlify(int(getattr(field, field_metadata.name)).to_bytes(field_metadata.length, 'big', signed=field_metadata.signed)).decode():<{2 *field_metadata.length + 1}}"
            for metadata in self.payload.metadata
            if metadata.type_ is list
            for field in getattr(self.payload, metadata.name)
            for field_metadata in field.metadata
        ]
        payload_values += [
            f" 0x{hexlify(getattr(self.payload, field.name)).decode():<{2 * self.payload.count + 1}}"
            for idx, field in enumerate(dataclasses.fields(self.payload)[1:])
            if self.payload.metadata[idx].type_ is bytes
            and hasattr(self.payload, "count")
        ]
        payload_values += [
            f" 0x{hexlify(getattr(self.payload, field.name)).decode():<{2 * self.payload.metadata[idx].length + 1}}"
            for idx, field in enumerate(dataclasses.fields(self.payload)[1:])
            if self.payload.metadata[idx].type_ is bytes
            and not hasattr(self.payload, "count")
        ]
        num_bytes = (
            sum(field.length for field in self.header.metadata)
            + 1
            + sum(field.length for field in self.payload.metadata)
        )
        num_bytes += sum(
            field_metadata.length
            for metadata in self.payload.metadata
            if metadata.type_ is list
            for field in getattr(self.payload, metadata.name)
            for field_metadata in field.metadata
        )
        num_bytes += sum(
            len(getattr(self.payload, field.name))
            for field in self.payload.metadata
            if field.type_ is bytes and field.length == 0
        )

        if self.payload_type not in [*PayloadType]:
            payload_type_str = "CUSTOM_DATA"
        else:
            payload_type_str = PayloadType(self.payload_type).name
        if num_bytes > 24:
            # put values on a separate row
            separators = header_separators + type_separators
            names = header_names + [" type "]
            values = header_values + type_value
            return (
                f" {' ' * 16}+{'+'.join(separators)}+\n"
                f" {payload_type_str:<16}|{'|'.join(names)}|\n"
                f" {f'({num_bytes} Bytes)':<16}|{'|'.join(values)}|\n"
                f" {' ' * 16}+{'+'.join(separators)}+\n"
                f" {' ' * 16}+{'+'.join(payload_separators)}+\n"
                f" {' ' * 16}|{'|'.join(payload_names)}|\n"
                f" {' ' * 16}|{'|'.join(payload_values)}|\n"
                f" {' ' * 16}+{'+'.join(payload_separators)}+\n"
            )

        # all in a row by default
        separators = header_separators + type_separators + payload_separators
        names = header_names + [" type "] + payload_names
        values = header_values + type_value + payload_values
        return (
            f" {' ' * 16}+{'+'.join(separators)}+\n"
            f" {payload_type_str:<16}|{'|'.join(names)}|\n"
            f" {f'({num_bytes} Bytes)':<16}|{'|'.join(values)}|\n"
            f" {' ' * 16}+{'+'.join(separators)}+\n"
        )



# SPDX-FileCopyrightText: 2022-present Inria
# SPDX-FileCopyrightText: 2022-present Alexandre Abadie <alexandre.abadie@inria.fr>
#
# SPDX-License-Identifier: BSD-3-Clause

"""Hacth custom hook module."""

# pylint: disable=import-error,too-few-public-methods,wrong-import-position

import os
import sys

from hatchling.builders.hooks.plugin.interface import BuildHookInterface

sys.path.append(os.path.dirname(__file__))
from pydotbot_utils import build_frontend, build_lh2  # noqa: E402


class CustomBuildHook(BuildHookInterface):
    """Custom build hook that will build the React web frontend."""

    def initialize(self, _, __):
        """Will be called before creating the source archive."""
        if os.getenv("SKIP_SDIST_HOOK") is not None:
            return
        build_frontend(self.root)
        build_lh2(self.root)



import dataclasses
from dataclasses import dataclass

import pytest

from dotbot.protocol import (
    PAYLOAD_PARSERS,
    ApplicationType,
    ControlModeType,
    Frame,
    Header,
    Packet,
    PacketFieldMetadata,
    PayloadAdvertisement,
    PayloadCommandMoveRaw,
    PayloadCommandRgbLed,
    PayloadCommandXgoAction,
    PayloadControlMode,
    PayloadDotBotData,
    PayloadDotBotSimulatorData,
    PayloadGPSPosition,
    PayloadGPSWaypoints,
    PayloadLH2Location,
    PayloadLh2RawData,
    PayloadLh2RawLocation,
    PayloadLH2Waypoints,
    PayloadRawData,
    PayloadSailBotData,
    PayloadType,
    ProtocolPayloadParserException,
    register_parser,
)


@dataclass
class PayloadWithBytesTest(Packet):

    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [
            PacketFieldMetadata(name="count", disp="len."),
            PacketFieldMetadata(name="data", type_=bytes, length=0),
        ]
    )
    count: int = 0
    data: bytes = b""


@dataclass
class PayloadWithBytesFixedLengthTest(Packet):

    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [
            PacketFieldMetadata(name="data", type_=bytes, length=8),
        ]
    )
    data: bytes = b""


register_parser(0x81, PayloadWithBytesTest)
register_parser(0x82, PayloadWithBytesFixedLengthTest)


@pytest.mark.parametrize(
    "bytes_,expected",
    [
        pytest.param(
            b"\x04\x02\x11\x11\x11\x11\x11\x22\x22\x11\x12\x12\x12\x12\x12\x12\x12\x12",
            Header(
                version=4,
                type_=2,
                destination=0x1122221111111111,
                source=0x1212121212121212,
            ),
            id="DefaultHeader",
        ),
    ],
)
def test_parse_header(bytes_, expected):
    assert Header().from_bytes(bytes_) == expected


@pytest.mark.parametrize(
    "bytes_,header,payload_type,payload",
    [
        pytest.param(
            b"\x04\x02\x11\x11\x11\x11\x11\x22\x22\x11\x12\x12\x12\x12\x12\x12\x12\x12\x04\x01",
            Header(
                version=4,
                type_=2,
                destination=0x1122221111111111,
                source=0x1212121212121212,
            ),
            PayloadType.ADVERTISEMENT,
            PayloadAdvertisement(application=ApplicationType.SailBot),
            id="PayloadAdvertisement",
        ),
        pytest.param(
            b"\x04\x02\x11\x11\x11\x11\x11\x22\x22\x11\x12\x12\x12\x12\x12\x12\x12\x12\x00\x00\x42\x00\x42",
            Header(
                version=4,
                type_=2,
                destination=0x1122221111111111,
                source=0x1212121212121212,
            ),
            PayloadType.CMD_MOVE_RAW,
            PayloadCommandMoveRaw(left_x=0, left_y=66, right_x=0, right_y=66),
            id="PayloadMoveRaw",
        ),
        pytest.param(
            b"\x04\x02\x11\x11\x11\x11\x11\x22\x22\x11\x12\x12\x12\x12\x12\x12\x12\x12\x01\x42\x42\x42",
            Header(
                version=4,
                type_=2,
                destination=0x1122221111111111,
                source=0x1212121212121212,
            ),
            PayloadType.CMD_RGB_LED,
            PayloadCommandRgbLed(red=66, green=66, blue=66),
            id="PayloadRgbLed",
        ),
        pytest.param(
            b"\x04\x02\x11\x11\x11\x11\x11\x22\x22\x11\x12\x12\x12\x12\x12\x12\x12\x12\x0b\x01",
            Header(
                version=4,
                type_=2,
                destination=0x1122221111111111,
                source=0x1212121212121212,
            ),
            PayloadType.CMD_XGO_ACTION,
            PayloadCommandXgoAction(action=1),
            id="PayloadCommandXgoAction",
        ),
        pytest.param(
            b"\x04\x02\x11\x11\x11\x11\x11\x22\x22\x11\x12\x12\x12\x12\x12\x12\x12\x12\x02"
            b"\x12\x34\x56\x78\x9a\xbc\xde\xf1\x01\x02",
            Header(
                version=4,
                type_=2,
                destination=0x1122221111111111,
                source=0x1212121212121212,
            ),
            PayloadType.LH2_RAW_LOCATION,
            PayloadLh2RawLocation(
                bits=0xF1DEBC9A78563412, polynomial_index=0x01, offset=0x02
            ),
            id="PayloadLH2RawLocation",
        ),
        pytest.param(
            b"\x04\x02\x88\x77\x66\x55\x44\x33\x22\x11\x21\x12\x22\x12\x22\x12\x22\x12\x0d"
            b"\x02"
            b"\x12\x34\x56\x78\x9a\xbc\xde\xf1\x01\x02"
            b"\x12\x34\x56\x78\x9a\xbc\xde\xf1\x02\x04",
            Header(
                version=4,
                type_=2,
                destination=0x1122334455667788,
                source=0x1222122212221221,
            ),
            PayloadType.LH2_RAW_DATA,
            PayloadLh2RawData(
                count=2,
                locations=[
                    PayloadLh2RawLocation(
                        bits=0xF1DEBC9A78563412, polynomial_index=0x01, offset=0x02
                    ),
                    PayloadLh2RawLocation(
                        bits=0xF1DEBC9A78563412, polynomial_index=0x02, offset=0x04
                    ),
                ],
            ),
            id="PayloadLH2RawData",
        ),
        pytest.param(
            b"\x04\x02\x11\x11\x11\x11\x11\x22\x22\x11\x12\x12\x12\x12\x12\x12\x12\x12\x03"
            b"\xe8\x03\x00\x00\xe8\x03\x00\x00\x02\x00\x00\x00",
            Header(
                version=4,
                type_=2,
                destination=0x1122221111111111,
                source=0x1212121212121212,
            ),
            PayloadType.LH2_LOCATION,
            PayloadLH2Location(pos_x=1000, pos_y=1000, pos_z=2),
            id="PayloadLH2Location",
        ),
        pytest.param(
            b"\x04\x02\x11\x11\x11\x11\x11\x22\x22\x11\x12\x12\x12\x12\x12\x12\x12\x12\x05"
            b"&~\xe9\x02]\xe4#\x00",
            Header(
                version=4,
                type_=2,
                destination=0x1122221111111111,
                source=0x1212121212121212,
            ),
            PayloadType.GPS_POSITION,
            PayloadGPSPosition(latitude=48856614, longitude=2352221),
            id="PayloadGPSPosition",
        ),
        pytest.param(
            b"\x04\x02\x88\x77\x66\x55\x44\x33\x22\x11\x21\x12\x22\x12\x22\x12\x22\x12\x0a"
            b"-\x00&~\xe9\x02]\xe4#\x00\xb4\x00\x1e\x14",
            Header(
                version=4,
                type_=2,
                destination=0x1122334455667788,
                source=0x1222122212221221,
            ),
            PayloadType.SAILBOT_DATA,
            PayloadSailBotData(
                direction=45,
                latitude=48856614,
                longitude=2352221,
                wind_angle=180,
                rudder_angle=30,
                sail_angle=20,
            ),
            id="PayloadSailbotData",
        ),
        pytest.param(
            b"\x04\x02\x88\x77\x66\x55\x44\x33\x22\x11\x21\x12\x22\x12\x22\x12\x22\x12\xfa"
            b"\x2d\x00"
            b"\x50\xc3\x00\x00"
            b"\xa8\x61\x00\x00",
            Header(
                version=4,
                type_=2,
                destination=0x1122334455667788,
                source=0x1222122212221221,
            ),
            PayloadType.DOTBOT_SIMULATOR_DATA,
            PayloadDotBotSimulatorData(theta=45, pos_x=50000, pos_y=25000),
            id="PayloadDotBotSimulatorData",
        ),
        pytest.param(
            b"\x04\x02\x11\x11\x11\x11\x11\x22\x22\x11\x12\x12\x12\x12\x12\x12\x12\x12\x07\x01",
            Header(
                version=4,
                type_=2,
                destination=0x1122221111111111,
                source=0x1212121212121212,
            ),
            PayloadType.CONTROL_MODE,
            PayloadControlMode(mode=ControlModeType.AUTO),
            id="PayloadControlMode",
        ),
        pytest.param(
            b"\x04\x02\x88\x77\x66\x55\x44\x33\x22\x11\x21\x12\x22\x12\x22\x12\x22\x12\x08"
            b"\x0a\x02"
            b"\xe8\x03\x00\x00\xe8\x03\x00\x00\x02\x00\x00\x00"
            b"\xe8\x03\x00\x00\xe8\x03\x00\x00\x02\x00\x00\x00",
            Header(
                version=4,
                type_=2,
                destination=0x1122334455667788,
                source=0x1222122212221221,
            ),
            PayloadType.LH2_WAYPOINTS,
            PayloadLH2Waypoints(
                threshold=10,
                count=2,
                waypoints=[
                    PayloadLH2Location(pos_x=1000, pos_y=1000, pos_z=2),
                    PayloadLH2Location(pos_x=1000, pos_y=1000, pos_z=2),
                ],
            ),
            id="PayloadLH2Waypoints",
        ),
        pytest.param(
            b"\x04\x02\x88\x77\x66\x55\x44\x33\x22\x11\x21\x12\x22\x12\x22\x12\x22\x12\x09"
            b"\x0a\x02"
            b"&~\xe9\x02]\xe4#\x00&~\xe9\x02]\xe4#\x00",
            Header(
                version=4,
                type_=2,
                destination=0x1122334455667788,
                source=0x1222122212221221,
            ),
            PayloadType.GPS_WAYPOINTS,
            PayloadGPSWaypoints(
                threshold=10,
                count=2,
                waypoints=[
                    PayloadGPSPosition(latitude=48856614, longitude=2352221),
                    PayloadGPSPosition(latitude=48856614, longitude=2352221),
                ],
            ),
            id="PayloadGPSWaypoints",
        ),
        pytest.param(
            b"\x01\x10\xff\xff\xff\xff\xff\xff\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00"  # header
            b"\x81"  # payload type
            b"\x08"  # count
            b"abcdefgh",  # data
            Header(),
            0x81,
            PayloadWithBytesTest(count=8, data=b"abcdefgh"),
            id="PayloadWithBytesTest",
        ),
        pytest.param(
            b"\x01\x10\xff\xff\xff\xff\xff\xff\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00"  # header
            b"\x82"  # payload type
            b"abcdefgh",  # data
            Header(),
            0x82,
            PayloadWithBytesFixedLengthTest(data=b"abcdefgh"),
            id="PayloadWithBytesFixedLengthTest",
        ),
        pytest.param(
            b"\x01\x10\xff\xff\xff\xff\xff\xff\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00"  # header
            b"\x10"  # Raw data type
            b"\x08"  # count
            b"abcdefgh",  # data
            Header(),
            0x10,
            PayloadRawData(count=8, data=b"abcdefgh"),
            id="PayloadRawDataTest",
        ),
    ],
)
def test_frame_parser(bytes_, header, payload_type, payload):
    frame = Frame().from_bytes(bytes_)
    assert frame.header == header
    assert frame.payload_type == payload_type
    assert frame.payload == payload


@pytest.mark.parametrize(
    "payload,expected",
    [
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                payload=PayloadCommandMoveRaw(
                    left_x=0, left_y=66, right_x=0, right_y=66
                ),
            ),
            b"\x04\x02\x88\x77\x66\x55\x44\x33\x22\x11\x21\x12\x22\x12\x22\x12\x22\x12\x00\x00\x42\x00\x42",
            id="PayloadMoveRaw1",
        ),
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadCommandMoveRaw(left_x=0, left_y=0, right_x=0, right_y=0),
            ),
            b"\x04\x02\x88\x77\x66\x55\x44\x33\x22\x11\x21\x12\x22\x12\x22\x12\x22\x12\x00\x00\x00\x00\x00",
            id="PayloadMoveRaw2",
        ),
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadCommandMoveRaw(left_x=-10, left_y=-10, right_x=-10, right_y=-10),
            ),
            b"\x04\x02\x88\x77\x66\x55\x44\x33\x22\x11\x21\x12\x22\x12\x22\x12\x22\x12\x00\xf6\xf6\xf6\xf6",
            id="PayloadMoveRaw3",
        ),
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadCommandRgbLed(red=0, green=0, blue=0),
            ),
            b"\x04\x02\x88\x77\x66\x55\x44\x33\x22\x11\x21\x12\x22\x12\x22\x12\x22\x12\x01\x00\x00\x00",
            id="PayloadRGBLed1",
        ),
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadCommandRgbLed(red=255, green=255, blue=255),
            ),
            b"\x04\x02\x88\x77\x66\x55\x44\x33\x22\x11\x21\x12\x22\x12\x22\x12\x22\x12\x01\xff\xff\xff",
            id="PayloadRGBLed2",
        ),
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadLh2RawData(
                    count=2,
                    locations=[
                        PayloadLh2RawLocation(
                            bits=0x123456789ABCDEF1, polynomial_index=0x01, offset=0x02
                        ),
                        PayloadLh2RawLocation(
                            bits=0x123456789ABCDEF1, polynomial_index=0x01, offset=0x02
                        ),
                    ],
                ),
            ),
            b"\x04\x02\x88\x77\x66\x55\x44\x33\x22\x11\x21\x12\x22\x12\x22\x12\x22\x12\x0d\x02"
            b"\xf1\xde\xbc\x9a\x78\x56\x34\x12\x01\x02"
            b"\xf1\xde\xbc\x9a\x78\x56\x34\x12\x01\x02",
            id="PayloadLH2RawData",
        ),
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadLH2Location(pos_x=1000, pos_y=1000, pos_z=2),
            ),
            b"\x04\x02\x88\x77\x66\x55\x44\x33\x22\x11\x21\x12\x22\x12\x22\x12\x22\x12\x03"
            b"\xe8\x03\x00\x00\xe8\x03\x00\x00\x02\x00\x00\x00",
            id="PayloadLH2Location",
        ),
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadAdvertisement(application=ApplicationType.SailBot),
            ),
            b"\x04\x02\x88\x77\x66\x55\x44\x33\x22\x11\x21\x12\x22\x12\x22\x12\x22\x12\x04\x01",
            id="PayloadAdvertisement",
        ),
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadGPSPosition(
                    latitude=48856614, longitude=2352221
                ),  # Paris coordinates
            ),
            b"\x04\x02\x88\x77\x66\x55\x44\x33\x22\x11\x21\x12\x22\x12\x22\x12\x22\x12\x05"
            b"&~\xe9\x02]\xe4#\x00",
            id="PayloadGPSPosition",
        ),
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadDotBotData(
                    direction=45,
                    count=2,
                    locations=[
                        PayloadLh2RawLocation(
                            bits=0x123456789ABCDEF1, polynomial_index=0x01, offset=0x02
                        ),
                        PayloadLh2RawLocation(
                            bits=0x123456789ABCDEF1, polynomial_index=0x01, offset=0x02
                        ),
                    ],
                ),
            ),
            b"\x04\x02\x88\x77\x66\x55\x44\x33\x22\x11\x21\x12\x22\x12\x22\x12\x22\x12\x06"
            b"-\x00\x02"
            b"\xf1\xde\xbc\x9a\x78\x56\x34\x12\x01\x02"
            b"\xf1\xde\xbc\x9a\x78\x56\x34\x12\x01\x02",
            id="PayloadDotBotData",
        ),
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadControlMode(mode=ControlModeType.AUTO),
            ),
            b"\x04\x02\x88\x77\x66\x55\x44\x33\x22\x11\x21\x12\x22\x12\x22\x12\x22\x12\x07\x01",
            id="PayloadControlMode",
        ),
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadLH2Waypoints(
                    threshold=10,
                    count=2,
                    waypoints=[
                        PayloadLH2Location(pos_x=1000, pos_y=1000, pos_z=2),
                        PayloadLH2Location(pos_x=1000, pos_y=1000, pos_z=2),
                    ],
                ),
            ),
            b"\x04\x02\x88\x77\x66\x55\x44\x33\x22\x11\x21\x12\x22\x12\x22\x12\x22\x12\x08\x0a\x02"
            b"\xe8\x03\x00\x00\xe8\x03\x00\x00\x02\x00\x00\x00"
            b"\xe8\x03\x00\x00\xe8\x03\x00\x00\x02\x00\x00\x00",
            id="PayloadLH2Waypoints",
        ),
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadGPSWaypoints(
                    threshold=10,
                    count=2,
                    waypoints=[
                        PayloadGPSPosition(latitude=48856614, longitude=2352221),
                        PayloadGPSPosition(latitude=48856614, longitude=2352221),
                    ],
                ),  # Paris coordinates x 2
            ),
            b"\x04\x02\x88\x77\x66\x55\x44\x33\x22\x11\x21\x12\x22\x12\x22\x12\x22\x12\x09\x0a\x02"
            b"&~\xe9\x02]\xe4#\x00&~\xe9\x02]\xe4#\x00",
            id="PayloadGPSWaypoints",
        ),
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadSailBotData(
                    direction=45,
                    latitude=48856614,
                    longitude=2352221,
                    wind_angle=180,
                    rudder_angle=30,
                    sail_angle=20,
                ),  # Paris coordinates
            ),
            b"\x04\x02\x88\x77\x66\x55\x44\x33\x22\x11\x21\x12\x22\x12\x22\x12\x22\x12\x0a"
            b"-\x00&~\xe9\x02]\xe4#\x00\xb4\x00\x1e\x14",
            id="PayloadSailBotData",
        ),
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadDotBotSimulatorData(
                    theta=45,
                    pos_x=50000,
                    pos_y=25000,
                ),
            ),
            b"\x04\x02\x88\x77\x66\x55\x44\x33\x22\x11\x21\x12\x22\x12\x22\x12\x22\x12\xfa"
            b"\x2d\x00"
            b"\x50\xc3\x00\x00"
            b"\xa8\x61\x00\x00",
            id="PayloadDotBotSimulatorData",
        ),
        pytest.param(
            Frame(
                header=Header(), payload=PayloadWithBytesTest(count=8, data=b"abcdefgh")
            ),
            b"\x01\x10\xff\xff\xff\xff\xff\xff\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00"  # header
            b"\x81"  # payload type
            b"\x08"  # count
            b"abcdefgh",  # data
            id="PayloadWithBytesTest",
        ),
        pytest.param(
            Frame(
                header=Header(),
                payload=PayloadWithBytesFixedLengthTest(data=b"abcdefgh"),
            ),
            b"\x01\x10\xff\xff\xff\xff\xff\xff\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00"  # header
            b"\x82"  # payload type
            b"abcdefgh",  # data
            id="PayloadWithBytesFixedLengthTest",
        ),
        pytest.param(
            Frame(
                header=Header(),
                payload=PayloadRawData(count=8, data=b"abcdefgh"),
            ),
            b"\x01\x10\xff\xff\xff\xff\xff\xff\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00"  # header
            b"\x10"  # Raw data type
            b"\x08"  # count
            b"abcdefgh",  # data
            id="PayloadRawDataTest",
        ),
    ],
)
def test_payload_to_bytes(payload, expected):
    result = payload.to_bytes()
    assert result == expected, f"{result} != {expected}"


@pytest.mark.parametrize(
    "payload,string",
    [
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadCommandMoveRaw(left_x=0, left_y=66, right_x=0, right_y=66),
            ),
            (
                "                 +------+------+--------------------+--------------------+------+------+------+------+------+\n"
                " CMD_MOVE_RAW    | ver. | type | dst                | src                | type | lx   | ly   | rx   | ry   |\n"
                " (23 Bytes)      | 0x04 | 0x02 | 0x1122334455667788 | 0x1222122212221221 | 0x00 | 0x00 | 0x42 | 0x00 | 0x42 |\n"
                "                 +------+------+--------------------+--------------------+------+------+------+------+------+\n"
                "\n"
            ),
            id="MoveRaw",
        ),
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadCommandRgbLed(red=0, green=0, blue=0),
            ),
            (
                "                 +------+------+--------------------+--------------------+------+------+------+------+\n"
                " CMD_RGB_LED     | ver. | type | dst                | src                | type | red  | green| blue |\n"
                " (22 Bytes)      | 0x04 | 0x02 | 0x1122334455667788 | 0x1222122212221221 | 0x01 | 0x00 | 0x00 | 0x00 |\n"
                "                 +------+------+--------------------+--------------------+------+------+------+------+\n"
                "\n"
            ),
            id="RGBLed",
        ),
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadLh2RawData(
                    count=2,
                    locations=[
                        PayloadLh2RawLocation(
                            bits=0x123456789ABCDEF1, polynomial_index=0x01, offset=0x02
                        ),
                        PayloadLh2RawLocation(
                            bits=0x123456789ABCDEF1, polynomial_index=0x01, offset=0x02
                        ),
                    ],
                ),
            ),
            (
                "                 +------+------+--------------------+--------------------+------+\n"
                " LH2_RAW_DATA    | ver. | type | dst                | src                | type |\n"
                " (40 Bytes)      | 0x04 | 0x02 | 0x1122334455667788 | 0x1222122212221221 | 0x0d |\n"
                "                 +------+------+--------------------+--------------------+------+\n"
                "                 +------+--------------------+------+------+--------------------+------+------+\n"
                "                 | len  | bits               | poly | off. | bits               | poly | off. |\n"
                "                 | 0x02 | 0x123456789abcdef1 | 0x01 | 0x02 | 0x123456789abcdef1 | 0x01 | 0x02 |\n"
                "                 +------+--------------------+------+------+--------------------+------+------+\n"
                "\n"
            ),
            id="LH2RawData",
        ),
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadLH2Location(pos_x=1000, pos_y=1000, pos_z=2),
            ),
            (
                "                 +------+------+--------------------+--------------------+------+\n"
                " LH2_LOCATION    | ver. | type | dst                | src                | type |\n"
                " (31 Bytes)      | 0x04 | 0x02 | 0x1122334455667788 | 0x1222122212221221 | 0x03 |\n"
                "                 +------+------+--------------------+--------------------+------+\n"
                "                 +------------+------------+------------+\n"
                "                 | x          | y          | z          |\n"
                "                 | 0x000003e8 | 0x000003e8 | 0x00000002 |\n"
                "                 +------------+------------+------------+\n"
                "\n"
            ),
            id="LH2Location",
        ),
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadAdvertisement(application=ApplicationType.SailBot),
            ),
            (
                "                 +------+------+--------------------+--------------------+------+------+\n"
                " ADVERTISEMENT   | ver. | type | dst                | src                | type | app  |\n"
                " (20 Bytes)      | 0x04 | 0x02 | 0x1122334455667788 | 0x1222122212221221 | 0x04 | 0x01 |\n"
                "                 +------+------+--------------------+--------------------+------+------+\n"
                "\n"
            ),
            id="Advertisement",
        ),
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadGPSPosition(
                    latitude=48856614, longitude=2352221
                ),  # Paris coordinates
            ),
            (
                "                 +------+------+--------------------+--------------------+------+\n"
                " GPS_POSITION    | ver. | type | dst                | src                | type |\n"
                " (27 Bytes)      | 0x04 | 0x02 | 0x1122334455667788 | 0x1222122212221221 | 0x05 |\n"
                "                 +------+------+--------------------+--------------------+------+\n"
                "                 +------------+------------+\n"
                "                 | lat.       | long.      |\n"
                "                 | 0x02e97e26 | 0x0023e45d |\n"
                "                 +------------+------------+\n"
                "\n"
            ),
            id="GPSPosition",
        ),
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadDotBotData(
                    direction=45,
                    count=2,
                    locations=[
                        PayloadLh2RawLocation(
                            bits=0x123456789ABCDEF1, polynomial_index=0x01, offset=0x02
                        ),
                        PayloadLh2RawLocation(
                            bits=0x123456789ABCDEF1, polynomial_index=0x01, offset=0x02
                        ),
                    ],
                ),
            ),
            (
                "                 +------+------+--------------------+--------------------+------+\n"
                " DOTBOT_DATA     | ver. | type | dst                | src                | type |\n"
                " (42 Bytes)      | 0x04 | 0x02 | 0x1122334455667788 | 0x1222122212221221 | 0x06 |\n"
                "                 +------+------+--------------------+--------------------+------+\n"
                "                 +--------+------+--------------------+------+------+--------------------+------+------+\n"
                "                 | dir.   | len  | bits               | poly | off. | bits               | poly | off. |\n"
                "                 | 0x002d | 0x02 | 0x123456789abcdef1 | 0x01 | 0x02 | 0x123456789abcdef1 | 0x01 | 0x02 |\n"
                "                 +--------+------+--------------------+------+------+--------------------+------+------+\n"
                "\n"
            ),
            id="DotBotData",
        ),
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadControlMode(mode=1),
            ),
            (
                "                 +------+------+--------------------+--------------------+------+------+\n"
                " CONTROL_MODE    | ver. | type | dst                | src                | type | mode |\n"
                " (20 Bytes)      | 0x04 | 0x02 | 0x1122334455667788 | 0x1222122212221221 | 0x07 | 0x01 |\n"
                "                 +------+------+--------------------+--------------------+------+------+\n"
                "\n"
            ),
            id="ControlMode",
        ),
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadLH2Waypoints(
                    threshold=10,
                    count=2,
                    waypoints=[
                        PayloadLH2Location(pos_x=1000, pos_y=1000, pos_z=2),
                        PayloadLH2Location(pos_x=1000, pos_y=1000, pos_z=2),
                    ],
                ),
            ),
            (
                "                 +------+------+--------------------+--------------------+------+\n"
                " LH2_WAYPOINTS   | ver. | type | dst                | src                | type |\n"
                " (45 Bytes)      | 0x04 | 0x02 | 0x1122334455667788 | 0x1222122212221221 | 0x08 |\n"
                "                 +------+------+--------------------+--------------------+------+\n"
                "                 +------+------+------------+------------+------------+------------+------------+------------+\n"
                "                 | thr. | len. | x          | y          | z          | x          | y          | z          |\n"
                "                 | 0x0a | 0x02 | 0x000003e8 | 0x000003e8 | 0x00000002 | 0x000003e8 | 0x000003e8 | 0x00000002 |\n"
                "                 +------+------+------------+------------+------------+------------+------------+------------+\n"
                "\n"
            ),
            id="LH2Waypoints",
        ),
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadGPSWaypoints(
                    threshold=10,
                    count=2,
                    waypoints=[
                        PayloadGPSPosition(latitude=48856614, longitude=2352221),
                        PayloadGPSPosition(latitude=48856614, longitude=2352221),
                    ],
                ),  # Paris coordinates x 2
            ),
            (
                "                 +------+------+--------------------+--------------------+------+\n"
                " GPS_WAYPOINTS   | ver. | type | dst                | src                | type |\n"
                " (37 Bytes)      | 0x04 | 0x02 | 0x1122334455667788 | 0x1222122212221221 | 0x09 |\n"
                "                 +------+------+--------------------+--------------------+------+\n"
                "                 +------+------+------------+------------+------------+------------+\n"
                "                 | thr. | len. | lat.       | long.      | lat.       | long.      |\n"
                "                 | 0x0a | 0x02 | 0x02e97e26 | 0x0023e45d | 0x02e97e26 | 0x0023e45d |\n"
                "                 +------+------+------------+------------+------------+------------+\n"
                "\n"
            ),
            id="GPSWaypoints",
        ),
        pytest.param(
            Frame(
                Header(
                    version=4,
                    type_=2,
                    destination=0x1122334455667788,
                    source=0x1222122212221221,
                ),
                PayloadSailBotData(
                    direction=45, latitude=48856614, longitude=2352221
                ),  # Paris coordinates
            ),
            (
                "                 +------+------+--------------------+--------------------+------+\n"
                " SAILBOT_DATA    | ver. | type | dst                | src                | type |\n"
                " (33 Bytes)      | 0x04 | 0x02 | 0x1122334455667788 | 0x1222122212221221 | 0x0a |\n"
                "                 +------+------+--------------------+--------------------+------+\n"
                "                 +--------+------------+------------+--------+------+------+\n"
                "                 | dir.   | lat.       | long.      | wind   | rud. | sail.|\n"
                "                 | 0x002d | 0x02e97e26 | 0x0023e45d | 0xffff | 0x00 | 0x00 |\n"
                "                 +--------+------------+------------+--------+------+------+\n"
                "\n"
            ),
            id="SailBotData",
        ),
        pytest.param(
            Frame(
                header=Header(),
                payload=PayloadWithBytesTest(count=8, data=b"abcdefgh"),
            ),
            (
                "                 +------+------+--------------------+--------------------+------+\n"
                " CUSTOM_DATA     | ver. | type | dst                | src                | type |\n"
                " (28 Bytes)      | 0x01 | 0x10 | 0xffffffffffffffff | 0x0000000000000000 | 0x81 |\n"
                "                 +------+------+--------------------+--------------------+------+\n"
                "                 +------+--------------------+\n"
                "                 | len. | data               |\n"
                "                 | 0x08 | 0x6162636465666768 |\n"
                "                 +------+--------------------+\n"
                "\n"
            ),
            id="PayloadWithBytesTest",
        ),
        pytest.param(
            Frame(
                header=Header(),
                payload=PayloadWithBytesFixedLengthTest(data=b"abcdefgh"),
            ),
            (
                "                 +------+------+--------------------+--------------------+------+\n"
                " CUSTOM_DATA     | ver. | type | dst                | src                | type |\n"
                " (27 Bytes)      | 0x01 | 0x10 | 0xffffffffffffffff | 0x0000000000000000 | 0x82 |\n"
                "                 +------+------+--------------------+--------------------+------+\n"
                "                 +--------------------+\n"
                "                 | data               |\n"
                "                 | 0x6162636465666768 |\n"
                "                 +--------------------+\n"
                "\n"
            ),
            id="PayloadWithBytesFixedLengthTest",
        ),
        pytest.param(
            Frame(
                header=Header(),
                payload=PayloadRawData(count=8, data=b"abcdefgh"),
            ),
            (
                "                 +------+------+--------------------+--------------------+------+\n"
                " RAW_DATA        | ver. | type | dst                | src                | type |\n"
                " (28 Bytes)      | 0x01 | 0x10 | 0xffffffffffffffff | 0x0000000000000000 | 0x10 |\n"
                "                 +------+------+--------------------+--------------------+------+\n"
                "                 +------+--------------------+\n"
                "                 | len. | data               |\n"
                "                 | 0x08 | 0x6162636465666768 |\n"
                "                 +------+--------------------+\n"
                "\n"
            ),
            id="PayloadRawDataTest",
        ),
    ],
)
def test_payload_frame_repr(payload, string, capsys):
    print(payload)
    out, _ = capsys.readouterr()
    assert out == string


def test_parse_missing_metadata():

    @dataclass
    class PayloadMissingMetadata(Packet):

        field: int = 0

    with pytest.raises(ValueError) as excinfo:
        PayloadMissingMetadata().from_bytes(b"")
    assert str(excinfo.value) == "metadata must be defined first"


@pytest.mark.parametrize(
    "packet,bytes_",
    [
        pytest.param(
            PayloadAdvertisement(application=ApplicationType.DotBot),
            b"",
            id="PayloadAdvertisement",
        ),
        pytest.param(
            PayloadDotBotData(
                direction=45,
                count=2,
                locations=[
                    PayloadLh2RawLocation(
                        bits=0x123456789ABCDEF1, polynomial_index=0x01, offset=0x02
                    ),
                    PayloadLh2RawLocation(
                        bits=0x123456789ABCDEF1, polynomial_index=0x01, offset=0x02
                    ),
                ],
            ),
            b"-\x00\x02" b"\xf1\xde\xbc\x9a\x78\x56\x34\x12\x01\x02",
            id="PayloadLh2RawLocation",
        ),
    ],
)
def test_from_bytes_empty(packet, bytes_):
    with pytest.raises(ValueError) as excinfo:
        packet.from_bytes(bytes_)
    assert str(excinfo.value) == "Not enough bytes to parse"


@dataclass
class PayloadTest(Packet):

    metadata: list[PacketFieldMetadata] = dataclasses.field(
        default_factory=lambda: [PacketFieldMetadata(name="field", type_=int)]
    )
    field: int = 0


@pytest.mark.parametrize(
    "payload_type,value_str",
    [
        (0x01, "0x01"),
        (0x0A, "0x0A"),
        (0xFA, "0xFA"),
    ],
)
def test_register_already_registered(payload_type, value_str):
    with pytest.raises(ValueError) as excinfo:
        register_parser(payload_type, PayloadTest)
    assert str(excinfo.value) == f"Payload type '{value_str}' already registered"


def test_register_reserved():
    with pytest.raises(ValueError) as excinfo:
        register_parser(0x7A, PayloadTest)
    assert str(excinfo.value) == "Payload type '0x7A' is reserved"


def test_register_parser():
    register_parser(0xFE, PayloadTest)
    assert PAYLOAD_PARSERS[0xFE] == PayloadTest


def test_parse_non_registered_payload():
    with pytest.raises(ProtocolPayloadParserException) as excinfo:
        Frame().from_bytes(
            b"\x04\x02\x88\x77\x66\x55\x44\x33\x22\x11\x21\x12\x22\x12\x22\x12\x22\x12\xfd\x01"
        )
    assert str(excinfo.value).startswith("Unsupported payload type")

    @dataclass
    class PayloadNotRegisteredTest(Packet):

        metadata: list[PacketFieldMetadata] = dataclasses.field(
            default_factory=lambda: [PacketFieldMetadata(name="field", type_=int)]
        )
        field: int = 0

    frame = Frame(header=Header(), payload=PayloadNotRegisteredTest())
    with pytest.raises(ValueError) as excinfo:
        frame.payload_type
    assert str(excinfo.value).startswith("Unsupported payload class")



# SPDX-FileCopyrightText: 2023-present Inria
# SPDX-FileCopyrightText: 2023-present Alexandre Abadie <alexandre.abadie@inria.fr>
#
# SPDX-License-Identifier: BSD-3-Clause

"""Python script used to analyze delays between LH2 packets received by the controller, sorted by DotBot."""

# pylint: disable=import-error,invalid-name,unspecified-encoding

import os
import sys
from datetime import datetime
from io import StringIO

import matplotlib.pyplot as plt
import numpy as np
from logfmt_pandas import read_logfmt

# Check the logs file path
log_filename = "pydotbot.log"
if len(sys.argv) > 1:
    log_path = os.path.join(sys.argv[1], log_filename)
else:
    log_path = os.path.join(os.path.dirname(__file__), "..", "..", log_filename)

if not os.path.exists(log_path):
    print(f"Logs file not found: '{log_path}'")
    sys.exit(1)

# Load the log file as Pandas dataframe
with open(log_path) as log:
    df = read_logfmt(StringIO(log.read()))
df = df[df["event"] == "lh2"].filter(items=["timestamp", "source"])
sources = df["source"].unique()

bins = [0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5, 10.5]
data = {}
print("======= Weights ========\n")
for source in sources:
    # Process the data into the update rate
    # ms / 100 (this helps make sure the pdf area calculation of the Histogram sums up to 100%)
    data[source] = (
        df[df["source"] == source]["timestamp"]
        .apply(lambda x: datetime.strptime(x, "%Y-%m-%dT%H:%M:%S.%fZ").timestamp())
        .diff()
        * 1000
        / 100
    )
    data[source].dropna(inplace=True)
    print(f"--- {source} ---")
    np_hist = np.histogram(data[source], bins, density=True)
    print(np_hist[0])
    print("------------------------\n")


# Plot the results
fig, ax = plt.subplots(nrows=1, ncols=1)
ax.hist(
    data.values(),
    bins=bins,
    density=True,
    histtype="bar",
    label=[f"{e} ({len(data[e])} samples)" for e in list(data.keys())],
)

# Add labels and grids
ax.grid()
ax.legend()
ax.set_xticks(
    list(range(11)),
    ["", "100", "200", "300", "400", "500", "600", "700", "800", "900", "1000"],
)
ax.set_yticks(
    [e / 10 for e in list(range(11))],
    ["", "10%", "20%", "30%", "40%", "50%", "60%", "70%", "80%", "90%", "100%"],
)

ax.set_xlabel("Time between packets [ms]")
ax.set_ylabel("Percentage of Packets [%]")
ax.set_title("Delay between LH2 packets")

plt.show()



"""Test module for the joystick main function."""

from unittest.mock import patch

from click.testing import CliRunner

from dotbot.joystick import main

MAIN_HELP_EXPECTED = """Usage: main [OPTIONS]

  DotBot joystick controller.

Options:
  -j, --joystick INTEGER          Index of the joystick to use. Defaults to 0
  -h, --hostname TEXT             Hostname of the controller. Defaults to
                                  'localhost'
  -p, --port INTEGER              HTTP port. Defaults to '8000'
  -s, --https                     Use HTTPS protocol instead of HTTP
  -d, --dotbot-address TEXT       Address in hex of the DotBot to control.
                                  Defaults to FFFFFFFFFFFFFFFF
  -a, --application [dotbot|sailbot]
                                  Application to control. Defaults to sailbot
  --log-level [debug|info|warning|error]
                                  Logging level. Defaults to info
  --help                          Show this message and exit.
"""


def test_joystick_main_help():
    runner = CliRunner()
    result = runner.invoke(main, ["--help"])
    assert result.exit_code == 0
    assert result.output == MAIN_HELP_EXPECTED


@patch("dotbot.version")
@patch("dotbot.joystick.JoystickController.start")
def test_main_no_joystick(start, version):
    version.return_value = "test"
    runner = CliRunner()
    result = runner.invoke(main)
    assert result.exit_code == 1
    assert "Error: No joystick connected." in result.output
    start.assert_not_called()



"""Test module for HDLC handler class."""

import pytest

from dotbot.hdlc import HDLCDecodeException, HDLCHandler, HDLCState


def test_hdlc_handler_states():
    handler = HDLCHandler()
    assert handler.state == HDLCState.IDLE
    handler.handle_byte(b"A")
    assert handler.state == HDLCState.IDLE
    handler.handle_byte(b"~")
    assert handler.state == HDLCState.RECEIVING
    handler.handle_byte(b"A")
    assert handler.state == HDLCState.RECEIVING
    handler.handle_byte(b"~")
    assert handler.state == HDLCState.READY
    handler.handle_byte(b"~")
    assert handler.state == HDLCState.RECEIVING
    handler.handle_byte(b"A")
    handler.handle_byte(b"\xf5")
    handler.handle_byte(b"\xa3")
    assert handler.state == HDLCState.RECEIVING
    handler.handle_byte(b"~")
    assert handler.state == HDLCState.READY
    handler.handle_byte(b"~")
    assert handler.state == HDLCState.RECEIVING
    handler.handle_byte(b"~")
    assert handler.output == bytearray()
    assert handler.state == HDLCState.RECEIVING
    handler.handle_byte(b"~")
    assert handler.output == bytearray()
    assert handler.state == HDLCState.RECEIVING


def test_hdlc_handler_decode():
    handler = HDLCHandler()
    for byte in b"~test\x88\x07~":
        handler.handle_byte(int(byte).to_bytes(1, "little"))
    assert handler.payload == b"test"
    assert handler.state == HDLCState.IDLE


def test_hdlc_handler_decode_with_flags():
    handler = HDLCHandler()
    for byte in b"~}^test}]\x06\x94~":
        handler.handle_byte(int(byte).to_bytes(1, "little"))
    assert handler.state == HDLCState.READY
    assert handler.payload == bytearray(b"~test}")
    assert handler.state == HDLCState.IDLE


def test_hdlc_handler_invalid_state():
    handler = HDLCHandler()
    for byte in b"~test\x42\x42":
        handler.handle_byte(int(byte).to_bytes(1, "little"))
    with pytest.raises(HDLCDecodeException) as exc:
        _ = handler.payload
    assert str(exc.value) == "Incomplete HDLC frame"


def test_hdlc_handler_invalid_fcs():
    handler = HDLCHandler()
    for byte in b"~test\x42\x42~":
        handler.handle_byte(int(byte).to_bytes(1, "little"))
    payload = handler.payload
    assert payload == bytearray()


def test_hdlc_handler_payload_too_short():
    handler = HDLCHandler()
    for byte in b"~a~":
        handler.handle_byte(int(byte).to_bytes(1, "little"))
    payload = handler.payload
    assert payload == bytearray()



# SPDX-FileCopyrightText: 2025-present Inria
# SPDX-FileCopyrightText: 2025-present Alexandre Abadie <alexandre.abadie@inria.fr>
#
# SPDX-License-Identifier: BSD-3-Clause

"""Interface of the Dotbot Edge Gateway."""

import asyncio
import base64
import time
import uuid
from binascii import hexlify
from dataclasses import dataclass
from typing import Dict

import serial
from gmqtt import Client as MQTTClient
from qrkey import qrkey_settings

from dotbot import DOTBOT_ADDRESS_DEFAULT, GATEWAY_ADDRESS_DEFAULT
from dotbot.hdlc import HDLCHandler, HDLCState, hdlc_encode
from dotbot.logger import LOGGER
from dotbot.models import DotBotModel
from dotbot.protocol import Frame, Header, PacketType, ProtocolPayloadParserException
from dotbot.serial_interface import SerialInterface, SerialInterfaceException

LOST_DELAY = 5  # seconds


@dataclass
class EdgeGatewaySettings:
    """Data class that holds controller settings."""

    port: str
    baudrate: int
    dotbot_address: str
    gw_address: str
    verbose: bool = False


class EdgeGateway:
    """Edge gateway class."""

    def __init__(self, settings: EdgeGatewaySettings):
        self.dotbots: Dict[str, DotBotModel] = {}
        self.logger = LOGGER.bind(context=__name__)
        self.header = Header(
            destination=int(DOTBOT_ADDRESS_DEFAULT, 16),
            source=int(settings.gw_address, 16),
        )
        self.settings = settings
        self.hdlc_handler = HDLCHandler()
        self.serial = None
        self.mqtt_client = None

    def on_connect(self, _, flags, rc, properties):
        logger = self.logger.bind(
            context=__name__,
            rc=rc,
            flags=flags,
            **properties,
        )
        logger.info("Edge Gateway connected to broker")
        self.mqtt_client.subscribe("/pydotbot/controller_to_edge")

    def on_message(self, _, topic, payload, qos, properties):
        """Called when a message is received from the controller."""
        logger = self.logger.bind(topic=topic)
        logger.info("MQTT message received", message=payload.decode())
        self.serial.write(hdlc_encode(base64.b64decode(payload)))

    def on_disconnect(self, _, packet, exc=None):
        logger = self.logger.bind(packet=packet, exc=exc)
        logger.info("Disconnected")

    def on_subscribe(self, client, mid, qos, properties):
        logger = self.logger.bind(qos=qos, **properties)
        topic = (
            client.get_subscriptions_by_mid(mid)[0].topic
            if client.get_subscriptions_by_mid(mid)
            else None
        )
        logger.info(f"Subscribed to {topic}")

    async def _start_serial(self):
        """Starts the serial listener thread in a coroutine."""
        self.logger.info("Starting serial interface")
        queue = asyncio.Queue()
        event_loop = asyncio.get_event_loop()

        def on_byte_received(byte):
            """Callback called on byte received."""
            event_loop.call_soon_threadsafe(queue.put_nowait, byte)

        self.serial = SerialInterface(
            self.settings.port, self.settings.baudrate, on_byte_received
        )
        await asyncio.sleep(1)
        self.serial.write(hdlc_encode(b"\x01\xff"))
        while 1:
            byte = await queue.get()
            self.handle_byte(byte)

    def handle_byte(self, byte):
        """Called on each byte received over UART."""
        self.hdlc_handler.handle_byte(byte)
        if self.hdlc_handler.state == HDLCState.READY:
            payload = self.hdlc_handler.payload
            if payload:
                try:
                    header = Header().from_bytes(payload)
                except ProtocolPayloadParserException:
                    self.logger.warning("Cannot parse header")
                    if self.settings.verbose is True:
                        print(header)
                    return
                source = hexlify(int(header.source).to_bytes(8, "big")).decode()
                logger = self.logger.bind(source=source)
                if header.type_ != PacketType.DATA:
                    logger.info("Skipping non data packet")
                    return
                if self.mqtt_client and self.mqtt_client.is_connected is False:
                    logger.warning("MQTT client not connected")
                    return
                if source == GATEWAY_ADDRESS_DEFAULT:
                    logger.warning("Invalid source in header")
                    return
                dotbot = DotBotModel(
                    address=source,
                    last_seen=time.time(),
                )
                self.dotbots.update({source: dotbot})
                logger.info(
                    "Forwarding serial payload to MQTT",
                    payload_type=hex(payload[header.size]),
                )
                self.mqtt_client.publish(
                    "/pydotbot/edge_to_controller",
                    base64.b64encode(payload).decode(),
                )

    def send_frame(self, frame: Frame):
        """Sends a command in an HDLC frame over serial."""
        destination = hexlify(int(frame.header.destination).to_bytes(8, "big")).decode()
        if destination not in self.dotbots:
            return
        if self.serial is not None:
            self.serial.write(hdlc_encode(frame.to_bytes()))
            self.logger.debug(
                "Payload sent",
                application=self.dotbots[destination].application.name,
                destination=destination,
                payload_type=frame.payload_type.name,
            )

    async def _connect_to_mqtt_broker(self):
        """Connect to the MQTT broker."""
        self.logger.info("Connecting to MQTT broker")
        self.mqtt_client: MQTTClient = MQTTClient(f"qrkey-{uuid.uuid4().hex}")
        self.mqtt_client.set_config(qrkey_settings.model_dump(exclude_none=True))
        self.mqtt_client.on_connect = self.on_connect
        self.mqtt_client.on_message = self.on_message
        self.mqtt_client.on_disconnect = self.on_disconnect
        self.mqtt_client.on_subscribe = self.on_subscribe
        await self.mqtt_client.connect(
            host=qrkey_settings.mqtt_host,
            port=qrkey_settings.mqtt_port,
            ssl=qrkey_settings.mqtt_use_ssl,
            keepalive=qrkey_settings.mqtt_keepalive,
            version=qrkey_settings.mqtt_version,
        )

    async def _dotbots_refresh(self):
        """Coroutine that periodically checks available dotbots."""
        while 1:
            dotbots_to_remove = []
            for address, dotbot in self.dotbots.items():
                if dotbot.last_seen + LOST_DELAY < time.time():
                    dotbots_to_remove.append(address)
            for address in dotbots_to_remove:
                self.logger.info("Removing lost dotbot", address=address)
                self.dotbots.pop(address)
            await asyncio.sleep(1)

    async def run(self):
        """Launch the controller."""
        tasks = []
        self.logger.info("Starting edge gateway")
        try:
            tasks = [
                asyncio.create_task(
                    name="MQTT broker client", coro=self._connect_to_mqtt_broker()
                ),
                # TODO better manage status of dotbots between the gateway board and the edge gateway
                # asyncio.create_task(
                #     name="Available dotbots check", coro=self._dotbots_refresh()
                # ),
                asyncio.create_task(name="Serial interface", coro=self._start_serial()),
            ]
            await asyncio.gather(*tasks)
        except (
            SerialInterfaceException,
            serial.serialutil.SerialException,
        ) as exc:
            self.logger.error(f"Error: {exc}")
        except SystemExit:
            pass
        finally:
            self.logger.info("Stopping edge gateway")
            self.serial.write(hdlc_encode(b"\x01\xfe"))
            for task in tasks:
                self.logger.info(f"Cancelling task '{task.get_name()}'")
                task.cancel()
            self.logger.info("Edge gateway stopped")



# SPDX-FileCopyrightText: 2022-present Inria
# SPDX-FileCopyrightText: 2022-present Alexandre Abadie <alexandre.abadie@inria.fr>
#
# SPDX-License-Identifier: BSD-3-Clause

"""Logger module."""

import logging
import logging.config

import structlog

LOG_LEVEL_MAP = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
}

SUPPORTED_HANDLERS_DEFAULT = {
    "console": {
        "formatter": "rich",
        "class": "logging.StreamHandler",
        "stream": "ext://sys.stderr",
    },
    "file": {
        "class": "logging.handlers.RotatingFileHandler",
        "formatter": "logfmt",
        "encoding": "utf-8",
    },
}


def setup_logging(filename, level, handlers):
    """Setup logging."""
    processors = [
        structlog.contextvars.merge_contextvars,
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        structlog.processors.TimeStamper(fmt="iso"),
        structlog.stdlib.PositionalArgumentsFormatter(),
        structlog.processors.StackInfoRenderer(),
        structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
    ]

    structlog.configure(
        processors=processors,
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )

    stdlib_handlers = {}
    for handler, value in SUPPORTED_HANDLERS_DEFAULT.items():
        if handler == "file":
            if filename is None:
                continue
            else:
                value["filename"] = filename
        stdlib_handlers.update({handler: value})

    stdlib_config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "logfmt": {
                "()": structlog.stdlib.ProcessorFormatter,
                "processor": structlog.processors.LogfmtRenderer(
                    key_order=["timestamp", "level", "logger", "event"],
                    drop_missing=True,
                ),
            },
            "rich": {
                "()": structlog.stdlib.ProcessorFormatter,
                "processor": structlog.dev.ConsoleRenderer(),
            },
        },
        "handlers": stdlib_handlers,
        "loggers": {
            "pydotbot": {
                "handlers": handlers,
                "level": LOG_LEVEL_MAP[level],
                "propagate": True,
            },
        },
    }
    logging.config.dictConfig(stdlib_config)


LOGGER = structlog.get_logger("pydotbot")



import asyncio
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from fastapi.testclient import TestClient
from httpx import ASGITransport, AsyncClient

from dotbot.models import (
    DotBotCalibrationStateModel,
    DotBotGPSPosition,
    DotBotLH2Position,
    DotBotModel,
    DotBotMoveRawCommandModel,
    DotBotRgbLedCommandModel,
)
from dotbot.protocol import (
    ApplicationType,
    Frame,
    Header,
    PayloadCommandMoveRaw,
    PayloadCommandRgbLed,
    PayloadGPSPosition,
    PayloadGPSWaypoints,
    PayloadLH2Location,
    PayloadLH2Waypoints,
)
from dotbot.server import api

client = AsyncClient(transport=ASGITransport(app=api), base_url="http://testserver")


@pytest.fixture(autouse=True)
def controller():
    api.controller = MagicMock()
    api.controller.websockets = []
    api.controller.header = MagicMock()
    api.controller.header.destination = MagicMock()
    api.controller.dotbots = MagicMock()
    api.controller.get_dotbots = MagicMock()
    api.controller.lh2_manager = MagicMock()
    api.controller.lh2_manager.state_model = DotBotCalibrationStateModel(state="test")
    api.controller.notify_clients = AsyncMock()
    api.controller.send_payload = MagicMock()
    api.controller.settings = MagicMock()
    api.controller.settings.gw_address = "0000"
    api.controller.settings.swarm_id = "0000"


@pytest.mark.asyncio
async def test_openapi_exists():
    response = await client.get("/api")
    assert response.status_code == 200


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "dotbots,code,found",
    [
        pytest.param(
            {
                "4242": DotBotModel(
                    address="4242",
                    application=ApplicationType.DotBot,
                    swarm="0000",
                    last_seen=123.4,
                ),
            },
            200,
            True,
            id="found",
        ),
        pytest.param(
            {
                "56789": DotBotModel(
                    address="56789",
                    application=ApplicationType.DotBot,
                    swarm="0000",
                    last_seen=123.4,
                ),
            },
            404,
            False,
            id="not_found",
        ),
    ],
)
async def test_set_dotbots_move_raw(dotbots, code, found):
    api.controller.dotbots = dotbots
    address = "4242"
    command = DotBotMoveRawCommandModel(left_x=42, left_y=0, right_x=42, right_y=0)
    header = Header(
        destination=int(address, 16),
        source=int(api.controller.settings.gw_address, 16),
    )
    payload = PayloadCommandMoveRaw(**command.model_dump())
    expected_frame = Frame(header=header, payload=payload)
    response = await client.put(
        f"/controller/dotbots/{address}/0/move_raw",
        json=command.model_dump(),
    )
    assert response.status_code == code
    if found is True:
        api.controller.send_payload.assert_called_with(expected_frame)
    else:
        api.controller.send_payload.assert_not_called()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "dotbots,code,found",
    [
        pytest.param(
            {
                "4242": DotBotModel(
                    address="4242",
                    application=ApplicationType.DotBot,
                    swarm="0000",
                    last_seen=123.4,
                ),
            },
            200,
            True,
            id="found",
        ),
        pytest.param(
            {
                "56789": DotBotModel(
                    address="56789",
                    application=ApplicationType.DotBot,
                    swarm="0000",
                    last_seen=123.4,
                ),
            },
            404,
            False,
            id="not_found",
        ),
    ],
)
async def test_set_dotbots_rgb_led(dotbots, code, found):
    api.controller.dotbots = dotbots
    address = "4242"
    command = DotBotRgbLedCommandModel(red=42, green=0, blue=42)
    header = Header(
        destination=int(address, 16),
        source=int(api.controller.settings.gw_address, 16),
    )
    payload = PayloadCommandRgbLed(**command.model_dump())
    expected_frame = Frame(header=header, payload=payload)
    response = await client.put(
        f"/controller/dotbots/{address}/0/rgb_led",
        json=command.model_dump(),
    )
    assert response.status_code == code

    if found:
        api.controller.send_payload.assert_called_with(expected_frame)
    else:
        api.controller.send_payload.assert_not_called()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "dotbots,has_position,application,message,code,found",
    [
        pytest.param(
            {
                "4242": DotBotModel(
                    address="4242",
                    application=ApplicationType.DotBot,
                    swarm="0000",
                    last_seen=123.4,
                ),
            },
            False,
            ApplicationType.DotBot,
            {"threshold": 10, "waypoints": [{"x": 0.5, "y": 0.1, "z": 0}]},
            200,
            True,
            id="dotbot_found",
        ),
        pytest.param(
            {
                "4242": DotBotModel(
                    address="4242",
                    application=ApplicationType.DotBot,
                    swarm="0000",
                    last_seen=123.4,
                    lh2_position=DotBotLH2Position(x=0.1, y=0.5, z=0),
                ),
            },
            True,
            ApplicationType.DotBot,
            {"threshold": 10, "waypoints": [{"x": 0.5, "y": 0.1, "z": 0}]},
            200,
            True,
            id="dotbot_with_position_found",
        ),
        pytest.param(
            {
                "56789": DotBotModel(
                    address="56789",
                    application=ApplicationType.DotBot,
                    swarm="0000",
                    last_seen=123.4,
                ),
            },
            False,
            ApplicationType.DotBot,
            {"threshold": 10, "waypoints": [{"x": 0.5, "y": 0.1, "z": 0}]},
            404,
            False,
            id="dotbot_not_found",
        ),
        pytest.param(
            {
                "4242": DotBotModel(
                    address="4242",
                    application=ApplicationType.SailBot,
                    swarm="0000",
                    last_seen=123.4,
                ),
            },
            False,
            ApplicationType.SailBot,
            {"threshold": 10, "waypoints": [{"latitude": 0.5, "longitude": 0.1}]},
            200,
            True,
            id="sailbot_found",
        ),
        pytest.param(
            {
                "4242": DotBotModel(
                    address="4242",
                    application=ApplicationType.SailBot,
                    swarm="0000",
                    last_seen=123.4,
                    gps_position=DotBotGPSPosition(latitude=0.1, longitude=0.5),
                ),
            },
            True,
            ApplicationType.SailBot,
            {"threshold": 10, "waypoints": [{"latitude": 0.5, "longitude": 0.1}]},
            200,
            True,
            id="sailbot_with_position_found",
        ),
        pytest.param(
            {
                "56789": DotBotModel(
                    address="56789",
                    application=ApplicationType.SailBot,
                    swarm="0000",
                    last_seen=123.4,
                ),
            },
            False,
            ApplicationType.SailBot,
            {"threshold": 10, "waypoints": [{"latitude": 0.5, "longitude": 0.1}]},
            404,
            False,
            id="sailbot_not_found",
        ),
    ],
)
async def test_set_dotbots_waypoints(
    dotbots, has_position, application, message, code, found
):
    api.controller.dotbots = dotbots
    address = "4242"
    header = Header(
        destination=int(address, 16),
        source=int(api.controller.settings.gw_address, 16),
    )
    if application == ApplicationType.SailBot:
        payload = PayloadGPSWaypoints(
            threshold=10,
            count=1,
            waypoints=[PayloadGPSPosition(latitude=500000, longitude=100000)],
        )
        expected_threshold = 10
        if has_position is True:
            expected_waypoints = [
                DotBotGPSPosition(latitude=0.1, longitude=0.5),
                DotBotGPSPosition(latitude=0.5, longitude=0.1),
            ]
        else:
            expected_waypoints = [DotBotGPSPosition(latitude=0.5, longitude=0.1)]
    else:  # DotBot application
        payload = PayloadLH2Waypoints(
            threshold=10,
            count=1,
            waypoints=[PayloadLH2Location(pos_x=500000, pos_y=100000, pos_z=0)],
        )
        expected_threshold = 10
        if has_position is True:
            expected_waypoints = [
                DotBotLH2Position(x=0.1, y=0.5, z=0),
                DotBotLH2Position(x=0.5, y=0.1, z=0),
            ]
        else:
            expected_waypoints = [DotBotLH2Position(x=0.5, y=0.1, z=0)]

    expected_frame = Frame(header=header, payload=payload)
    response = await client.put(
        f"/controller/dotbots/{address}/{application.value}/waypoints",
        json=message,
    )
    assert response.status_code == code

    if found:
        api.controller.send_payload.assert_called_with(expected_frame)
        assert api.controller.dotbots[address].waypoints == expected_waypoints
        assert api.controller.dotbots[address].waypoints_threshold == expected_threshold
    else:
        api.controller.send_payload.assert_not_called()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "dotbots,result",
    [
        pytest.param({}, [], id="empty"),
        pytest.param(
            {
                "12345": DotBotModel(
                    address="12345",
                    application=ApplicationType.DotBot,
                    swarm="0000",
                    last_seen=123.4,
                ),
            },
            [
                DotBotModel(
                    address="12345",
                    application=ApplicationType.DotBot,
                    swarm="0000",
                    last_seen=123.4,
                ).model_dump(exclude_none=True),
            ],
            id="one",
        ),
        pytest.param(
            {
                "56789": DotBotModel(
                    address="56789",
                    application=ApplicationType.DotBot,
                    swarm="0000",
                    last_seen=123.4,
                ),
                "12345": DotBotModel(
                    address="12345",
                    application=ApplicationType.DotBot,
                    swarm="0000",
                    last_seen=123.4,
                ),
            },
            [
                DotBotModel(
                    address="12345",
                    application=ApplicationType.DotBot,
                    swarm="0000",
                    last_seen=123.4,
                ).model_dump(exclude_none=True),
                DotBotModel(
                    address="56789",
                    application=ApplicationType.DotBot,
                    swarm="0000",
                    last_seen=123.4,
                ).model_dump(exclude_none=True),
            ],
            id="sorted",
        ),
    ],
)
async def test_get_dotbots(dotbots, result):
    api.controller.get_dotbots.return_value = list(
        sorted(dotbots.values(), key=lambda dotbot: dotbot.address)
    )
    response = await client.get("/controller/dotbots")
    assert response.status_code == 200
    assert response.json() == result


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "dotbots,address,code,found,result",
    [
        pytest.param(
            {
                "56789": DotBotModel(
                    address="56789",
                    application=ApplicationType.DotBot,
                    swarm="0000",
                    last_seen=123.4,
                ),
                "12345": DotBotModel(
                    address="12345",
                    application=ApplicationType.DotBot,
                    swarm="0000",
                    last_seen=123.4,
                ),
            },
            "12345",
            200,
            True,
            DotBotModel(
                address="12345",
                application=ApplicationType.DotBot,
                swarm="0000",
                last_seen=123.4,
            ).model_dump(exclude_none=True),
            id="found",
        ),
        pytest.param(
            {
                "56789": DotBotModel(
                    address="56789",
                    application=ApplicationType.DotBot,
                    swarm="0000",
                    last_seen=123.4,
                ),
                "12345": DotBotModel(
                    address="12345",
                    application=ApplicationType.DotBot,
                    swarm="0000",
                    last_seen=123.4,
                ),
            },
            "34567",
            404,
            False,
            None,
            id="not_found",
        ),
    ],
)
async def test_get_dotbot(dotbots, address, code, found, result):
    api.controller.dotbots = dotbots
    response = await client.get(f"/controller/dotbots/{address}")
    assert response.status_code == code
    if found is True:
        assert response.json() == result


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "dotbots,address,code,found",
    [
        pytest.param(
            {
                "56789": DotBotModel(
                    address="56789",
                    application=ApplicationType.DotBot,
                    swarm="0000",
                    last_seen=123.4,
                    position_history=[
                        DotBotLH2Position(x=0.0, y=0.5, z=0.0),
                        DotBotLH2Position(x=0.5, y=0.5, z=0.0),
                    ],
                ),
                "12345": DotBotModel(
                    address="12345",
                    application=ApplicationType.DotBot,
                    swarm="0000",
                    last_seen=123.4,
                    position_history=[
                        DotBotLH2Position(x=0.5, y=0.5, z=0.0),
                        DotBotLH2Position(x=0.5, y=0.0, z=0.0),
                    ],
                ),
            },
            "12345",
            200,
            True,
            id="dotbot_found",
        ),
        pytest.param(
            {
                "56789": DotBotModel(
                    address="56789",
                    application=ApplicationType.DotBot,
                    swarm="0000",
                    last_seen=123.4,
                ),
                "12345": DotBotModel(
                    address="12345",
                    application=ApplicationType.DotBot,
                    swarm="0000",
                    last_seen=123.4,
                ),
            },
            "34567",
            404,
            False,
            id="dotbot_not_found",
        ),
        pytest.param(
            {
                "56789": DotBotModel(
                    address="56789",
                    application=ApplicationType.SailBot,
                    swarm="0000",
                    last_seen=123.4,
                    position_history=[
                        DotBotGPSPosition(latitude=45.7597, longitude=4.8422),
                        DotBotGPSPosition(latitude=48.8567, longitude=2.3508),
                    ],
                ),
                "12345": DotBotModel(
                    address="12345",
                    application=ApplicationType.SailBot,
                    swarm="0000",
                    last_seen=123.4,
                    position_history=[
                        DotBotGPSPosition(latitude=51.509865, longitude=-0.118092),
                        DotBotGPSPosition(latitude=48.8567, longitude=2.3508),
                    ],
                ),
            },
            "34567",
            404,
            False,
            id="sailbot_found",
        ),
        pytest.param(
            {
                "56789": DotBotModel(
                    address="56789",
                    application=ApplicationType.SailBot,
                    swarm="0000",
                    last_seen=123.4,
                ),
                "12345": DotBotModel(
                    address="12345",
                    application=ApplicationType.SailBot,
                    swarm="0000",
                    last_seen=123.4,
                ),
            },
            "34567",
            404,
            False,
            id="sailbot_not_found",
        ),
    ],
)
async def test_clear_dotbot_position_history(dotbots, address, code, found):
    api.controller.dotbots = dotbots
    response = await client.delete(f"/controller/dotbots/{address}/positions")
    assert response.status_code == code
    if found is True:
        assert api.controller.dotbots[address].position_history == []


@pytest.mark.asyncio
async def test_lh2_calibration():
    response = await client.get("/controller/lh2/calibration")
    assert response.json() == DotBotCalibrationStateModel(state="test").model_dump()
    assert response.status_code == 200

    with patch(
        "dotbot.server.api.controller.lh2_manager.add_calibration_point"
    ) as point:
        response = await client.post(
            "/controller/lh2/calibration/2",
        )
        assert response.status_code == 200
        point.assert_called_with(2)

    with patch(
        "dotbot.server.api.controller.lh2_manager.compute_calibration"
    ) as calibration:
        response = await client.put(
            "/controller/lh2/calibration",
        )
        assert response.status_code == 200
        calibration.assert_called_once()


@pytest.mark.asyncio
async def test_ws_client():
    with TestClient(api).websocket_connect("/controller/ws/status") as websocket:
        await asyncio.sleep(0.1)
        assert len(api.controller.websockets) == 1
        websocket.close()
        await asyncio.sleep(0.1)
        assert len(api.controller.websockets) == 0


# @pytest.mark.asyncio
# @patch("uvicorn.Server.serve")
# async def test_web(serve, caplog):
#     caplog.set_level(logging.DEBUG, logger="pydotbot")
#     with pytest.raises(SystemExit):
#         await web(None)
#     serve.side_effect = asyncio.exceptions.CancelledError()
#     await web(None)
#     assert "Web server cancelled" in caplog.text



# SPDX-FileCopyrightText: 2022-present Inria
# SPDX-FileCopyrightText: 2022-present Alexandre Abadie <alexandre.abadie@inria.fr>
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module implementing HDLC protocol primitives."""

from enum import Enum

from dotbot.logger import LOGGER

HDLC_FLAG = b"\x7e"
HDLC_FLAG_ESCAPED = b"\x5e"
HDLC_ESCAPE = b"\x7d"
HDLC_ESCAPE_ESCAPED = b"\x5d"
HDLC_FCS_INIT = 0xFFFF
HDLC_FCS_OK = 0xF0B8

# fmt: off
FCS16TAB = (
    0x0000, 0x1189, 0x2312, 0x329B, 0x4624, 0x57AD, 0x6536, 0x74BF,
    0x8C48, 0x9DC1, 0xAF5A, 0xBED3, 0xCA6C, 0xDBE5, 0xE97E, 0xF8F7,
    0x1081, 0x0108, 0x3393, 0x221A, 0x56A5, 0x472C, 0x75B7, 0x643E,
    0x9CC9, 0x8D40, 0xBFDB, 0xAE52, 0xDAED, 0xCB64, 0xF9FF, 0xE876,
    0x2102, 0x308B, 0x0210, 0x1399, 0x6726, 0x76AF, 0x4434, 0x55BD,
    0xAD4A, 0xBCC3, 0x8E58, 0x9FD1, 0xEB6E, 0xFAE7, 0xC87C, 0xD9F5,
    0x3183, 0x200A, 0x1291, 0x0318, 0x77A7, 0x662E, 0x54B5, 0x453C,
    0xBDCB, 0xAC42, 0x9ED9, 0x8F50, 0xFBEF, 0xEA66, 0xD8FD, 0xC974,
    0x4204, 0x538D, 0x6116, 0x709F, 0x0420, 0x15A9, 0x2732, 0x36BB,
    0xCE4C, 0xDFC5, 0xED5E, 0xFCD7, 0x8868, 0x99E1, 0xAB7A, 0xBAF3,
    0x5285, 0x430C, 0x7197, 0x601E, 0x14A1, 0x0528, 0x37B3, 0x263A,
    0xDECD, 0xCF44, 0xFDDF, 0xEC56, 0x98E9, 0x8960, 0xBBFB, 0xAA72,
    0x6306, 0x728F, 0x4014, 0x519D, 0x2522, 0x34AB, 0x0630, 0x17B9,
    0xEF4E, 0xFEC7, 0xCC5C, 0xDDD5, 0xA96A, 0xB8E3, 0x8A78, 0x9BF1,
    0x7387, 0x620E, 0x5095, 0x411C, 0x35A3, 0x242A, 0x16B1, 0x0738,
    0xFFCF, 0xEE46, 0xDCDD, 0xCD54, 0xB9EB, 0xA862, 0x9AF9, 0x8B70,
    0x8408, 0x9581, 0xA71A, 0xB693, 0xC22C, 0xD3A5, 0xE13E, 0xF0B7,
    0x0840, 0x19C9, 0x2B52, 0x3ADB, 0x4E64, 0x5FED, 0x6D76, 0x7CFF,
    0x9489, 0x8500, 0xB79B, 0xA612, 0xD2AD, 0xC324, 0xF1BF, 0xE036,
    0x18C1, 0x0948, 0x3BD3, 0x2A5A, 0x5EE5, 0x4F6C, 0x7DF7, 0x6C7E,
    0xA50A, 0xB483, 0x8618, 0x9791, 0xE32E, 0xF2A7, 0xC03C, 0xD1B5,
    0x2942, 0x38CB, 0x0A50, 0x1BD9, 0x6F66, 0x7EEF, 0x4C74, 0x5DFD,
    0xB58B, 0xA402, 0x9699, 0x8710, 0xF3AF, 0xE226, 0xD0BD, 0xC134,
    0x39C3, 0x284A, 0x1AD1, 0x0B58, 0x7FE7, 0x6E6E, 0x5CF5, 0x4D7C,
    0xC60C, 0xD785, 0xE51E, 0xF497, 0x8028, 0x91A1, 0xA33A, 0xB2B3,
    0x4A44, 0x5BCD, 0x6956, 0x78DF, 0x0C60, 0x1DE9, 0x2F72, 0x3EFB,
    0xD68D, 0xC704, 0xF59F, 0xE416, 0x90A9, 0x8120, 0xB3BB, 0xA232,
    0x5AC5, 0x4B4C, 0x79D7, 0x685E, 0x1CE1, 0x0D68, 0x3FF3, 0x2E7A,
    0xE70E, 0xF687, 0xC41C, 0xD595, 0xA12A, 0xB0A3, 0x8238, 0x93B1,
    0x6B46, 0x7ACF, 0x4854, 0x59DD, 0x2D62, 0x3CEB, 0x0E70, 0x1FF9,
    0xF78F, 0xE606, 0xD49D, 0xC514, 0xB1AB, 0xA022, 0x92B9, 0x8330,
    0x7BC7, 0x6A4E, 0x58D5, 0x495C, 0x3DE3, 0x2C6A, 0x1EF1, 0x0F78,
)
# fmt: on


class HDLCDecodeException(Exception):
    """Exception raised when decoding wrong HDLC frames."""


def _fcs_update(fcs, byte):
    return (fcs >> 8) ^ FCS16TAB[((fcs ^ ord(byte)) & 0xFF)]


def _to_byte(value):
    return int(value).to_bytes(1, "little")


def _escape_byte(byte) -> bytes:
    result = bytearray()
    if byte == HDLC_ESCAPE:
        result += HDLC_ESCAPE
        result += HDLC_ESCAPE_ESCAPED
    elif byte == HDLC_FLAG:
        result += HDLC_ESCAPE
        result += HDLC_FLAG_ESCAPED
    else:
        result += byte
    return result


def hdlc_encode(payload: bytes) -> bytes:
    """Encodes a payload in an HDLC frame.
    >>> hdlc_encode(b"test")
    bytearray(b'~test\\x88\\x07~')
    >>> hdlc_encode(b"")
    bytearray(b'~\\x00\\x00~')
    >>> hdlc_encode(b"\\x00\\x00\\xf6\\xf6\\xf6\\xf6")
    bytearray(b'~\\x00\\x00\\xf6\\xf6\\xf6\\xf6\\xb2+~')
    >>> hdlc_encode(b"\\x00\\x01\\n\\n\\n")
    bytearray(b'~\\x00\\x01\\n\\n\\n\\x9c\\xf2~')
    >>> hdlc_encode(b"~test~")
    bytearray(b'~}^test}^\\x9d\\xa6~')
    >>> hdlc_encode(b"~test}")
    bytearray(b'~}^test}]\\x06\\x94~')
    >>> hdlc_encode(b"\\xe7\\x94:\\xa6")
    bytearray(b'~\\xe7\\x94:\\xa6\\x83}^~')
    >>> hdlc_encode(b"'$W\\x82")
    bytearray(b"~\\'$W\\x82\\x13}]~")
    """
    # initialize output buffer
    hdlc_frame = bytearray()

    # initialize frame check sequence
    fcs = HDLC_FCS_INIT

    # add start flag
    hdlc_frame += HDLC_FLAG

    # write payload in frame
    for byte in payload:
        fcs = _fcs_update(fcs, _to_byte(byte))
        hdlc_frame += _escape_byte(_to_byte(byte))
    fcs = 0xFFFF - fcs

    # add FCS
    hdlc_frame += _escape_byte(_to_byte(fcs & 0xFF))
    hdlc_frame += _escape_byte(_to_byte((fcs & 0xFF00) >> 8))

    # add end flag
    hdlc_frame += HDLC_FLAG

    return hdlc_frame


def hdlc_decode(frame: bytes) -> bytes:
    """Decodes an HDLC frame and return the payload it contains.

    >>> hdlc_decode(b"~test\\x88\\x07~")
    bytearray(b'test')
    >>> hdlc_decode(b"~\\x00\\x00\\xf6\\xf6\\xf6\\xf6\\xb2+~")
    bytearray(b'\\x00\\x00\\xf6\\xf6\\xf6\\xf6')
    >>> hdlc_decode(b"~\\x00\\x01\\n\\n\\n\\x9c\\xf2~")
    bytearray(b'\\x00\\x01\\n\\n\\n')
    >>> hdlc_decode(b"~}^test}^\\x9d\\xa6~")
    bytearray(b'~test~')
    >>> hdlc_decode(b"~}^test}]\\x06\\x94~")
    bytearray(b'~test}')
    >>> hdlc_decode(b"~\\xe7\\x94:\\xa6\\x83}^~")
    bytearray(b'\\xe7\\x94:\\xa6')
    >>> hdlc_decode(b"~\\'$W\\x82\\x13}]~")
    bytearray(b"\\'$W\\x82")
    >>> hdlc_decode(b"~\\x00\\x00~")
    bytearray(b'')
    >>> hdlc_decode(b"~test\\x42\\x42~")
    Traceback (most recent call last):
    dotbot.hdlc.HDLCDecodeException: Invalid FCS
    >>> hdlc_decode(b"~\\x00~")
    Traceback (most recent call last):
    dotbot.hdlc.HDLCDecodeException: Invalid payload
    """
    output = bytearray()
    fcs = HDLC_FCS_INIT
    escape_byte = False
    for byte in frame[1:-1]:
        byte = _to_byte(byte)
        if byte == HDLC_ESCAPE:
            escape_byte = True
        elif escape_byte is True:
            if byte == HDLC_ESCAPE_ESCAPED:
                output += HDLC_ESCAPE
                fcs = _fcs_update(fcs, HDLC_ESCAPE)
            elif byte == HDLC_FLAG_ESCAPED:
                output += HDLC_FLAG
                fcs = _fcs_update(fcs, HDLC_FLAG)
            escape_byte = False
        else:
            output += byte
            fcs = _fcs_update(fcs, byte)
    if len(output) < 2:
        raise HDLCDecodeException("Invalid payload")
    if fcs != HDLC_FCS_OK:
        raise HDLCDecodeException("Invalid FCS")
    return output[:-2]


class HDLCState(Enum):
    """State of the HDLC handler."""

    IDLE = 0
    RECEIVING = 1
    READY = 2


class HDLCHandler:
    """Handles the reception of an HDLC frame byte by byte."""

    def __init__(self):
        self.state = HDLCState.IDLE
        self.fcs = HDLC_FCS_INIT
        self.output = bytearray()
        self.escape_byte = False
        self._logger = LOGGER.bind(context=__name__)

    @property
    def payload(self):
        """Returns the payload contained in a frame."""
        if self.state != HDLCState.READY:
            raise HDLCDecodeException("Incomplete HDLC frame")

        self.state = HDLCState.IDLE
        if len(self.output) < 2:
            self._logger.error("Invalid payload")
            return bytearray()
        if self.fcs != HDLC_FCS_OK:
            self._logger.error("Invalid FCS")
            return bytearray()
        self.fcs = HDLC_FCS_INIT
        return self.output[:-2]

    def handle_byte(self, byte):
        """Handle new byte received."""
        if self.state in [HDLCState.IDLE, HDLCState.READY] and byte == HDLC_FLAG:
            self.output = bytearray()
            self.fcs = HDLC_FCS_INIT
            self.state = HDLCState.RECEIVING
        elif self.output and self.state == HDLCState.RECEIVING and byte == HDLC_FLAG:
            # End of frame
            self.state = HDLCState.READY
        elif self.state == HDLCState.RECEIVING and byte != HDLC_FLAG:
            # Middle of the frame
            if byte == HDLC_ESCAPE:
                self.escape_byte = True
            elif self.escape_byte is True:
                if byte == HDLC_ESCAPE_ESCAPED:
                    self.output += HDLC_ESCAPE
                    self.fcs = _fcs_update(self.fcs, HDLC_ESCAPE)
                elif byte == HDLC_FLAG_ESCAPED:
                    self.output += HDLC_FLAG
                    self.fcs = _fcs_update(self.fcs, HDLC_FLAG)
                self.escape_byte = False
            else:
                self.output += byte
                self.fcs = _fcs_update(self.fcs, byte)



"""Test module for controller base class."""

import asyncio
import time
from unittest.mock import patch

import pytest
import serial

from dotbot.controller import Controller, ControllerSettings, gps_distance, lh2_distance
from dotbot.hdlc import hdlc_encode
from dotbot.models import DotBotGPSPosition, DotBotLH2Position, DotBotModel
from dotbot.protocol import ControlModeType, Frame, Header, PayloadControlMode


@pytest.mark.asyncio
@patch("dotbot.serial_interface.serial.Serial.write")
@patch("dotbot.serial_interface.serial.Serial.open")
@patch("dotbot.serial_interface.serial.Serial.flush")
async def test_controller(_, __, serial_write, capsys):
    """Check controller subclass instanciation and write to serial."""
    settings = ControllerSettings("/dev/null", "115200", "0", "456", "78")
    controller = Controller(settings)
    controller.dotbots.update(
        {
            "0000000000000000": DotBotModel(
                address="0000000000000000", last_seen=time.time()
            )
        }
    )
    controller.serial = serial.Serial(settings.port, settings.baudrate)
    frame = Frame(
        header=Header(
            destination=0,
            source=0,
        ),
        payload=PayloadControlMode(mode=ControlModeType.AUTO),
    )
    controller.send_payload(frame)
    assert serial_write.call_count == 1
    payload_expected = hdlc_encode(frame.to_bytes())
    assert serial_write.call_args_list[0].args[0] == payload_expected


@pytest.mark.asyncio
@patch("dotbot.serial_interface.serial.Serial.write")
@patch("dotbot.serial_interface.serial.Serial.open")
@patch("dotbot.serial_interface.serial.Serial.flush")
async def test_controller_dont_send(_, __, serial_write):
    """Check controller subclass instanciation and write to serial."""
    settings = ControllerSettings("/dev/null", "115200", "0", "456", "78")
    controller = Controller(settings)
    dotbot = DotBotModel(address="0000000000000000", last_seen=time.time())
    controller.dotbots.update({dotbot.address: dotbot})
    controller.serial = serial.Serial(settings.port, settings.baudrate)
    frame = Frame(
        header=Header(
            destination=1,
            source=0,
        ),
        payload=PayloadControlMode(mode=ControlModeType.AUTO),
    )
    # DotBot is not in the controller known dotbot, so the payload won't be sent
    controller.send_payload(frame)
    assert serial_write.call_count == 0


def test_controller_saibot_simulator():
    """Check controller called for sailbot simulator."""

    async def start_simulator():
        settings = ControllerSettings(
            "sailbot-simulator", "115200", "0", "456", "78", 8000
        )
        controller = Controller(settings)
        try:
            await asyncio.wait_for(controller.run(), timeout=0.5)
        except asyncio.TimeoutError:
            pass

    loop = asyncio.get_event_loop()
    loop.run_until_complete(start_simulator())


@patch("dotbot.controller.QrkeyController.start")
def test_controller_dotbot_simulator(_):
    """Check controller called for dotbot simulator."""

    async def start_simulator():
        settings = ControllerSettings(
            "dotbot-simulator", "115200", "0", "456", "78", 8001
        )
        controller = Controller(settings)
        try:
            await asyncio.wait_for(controller.run(), timeout=0.5)
        except asyncio.TimeoutError:
            pass

    loop = asyncio.get_event_loop()
    loop.run_until_complete(start_simulator())


@pytest.mark.parametrize(
    "last,new,result",
    [
        (DotBotLH2Position(x=0, y=0, z=0), DotBotLH2Position(x=0, y=0, z=0), 0.0),
        (DotBotLH2Position(x=1, y=0, z=0), DotBotLH2Position(x=0, y=0, z=0), 1.0),
        (DotBotLH2Position(x=0, y=1, z=0), DotBotLH2Position(x=0, y=0, z=0), 1.0),
    ],
)
def test_lh2_distance(last, new, result):
    assert lh2_distance(last, new) == result


@pytest.mark.parametrize(
    "last,new,result",
    [
        (
            DotBotGPSPosition(latitude=45.7597, longitude=4.8422),  # Lyon
            DotBotGPSPosition(latitude=48.8567, longitude=2.3508),  # Paris
            392217.25594,
        ),
        (
            DotBotGPSPosition(latitude=51.509865, longitude=-0.118092),  # London
            DotBotGPSPosition(latitude=48.8567, longitude=2.3508),  # Paris
            343374.55271,
        ),
    ],
)
def test_gps_distance(last, new, result):
    assert gps_distance(last, new) == pytest.approx(result)



# SPDX-FileCopyrightText: 2022-present Inria
# SPDX-FileCopyrightText: 2022-present Alexandre Abadie <alexandre.abadie@inria.fr>
#
# SPDX-License-Identifier: BSD-3-Clause

#!/usr/bin/env python3

"""Main module of the Dotbot controller command line tool."""

import asyncio
import os
import sys

import click
import serial

from dotbot import (
    CONTROLLER_PORT_DEFAULT,
    DOTBOT_ADDRESS_DEFAULT,
    GATEWAY_ADDRESS_DEFAULT,
    SERIAL_BAUDRATE_DEFAULT,
    SERIAL_PORT_DEFAULT,
    SWARM_ID_DEFAULT,
    pydotbot_version,
)
from dotbot.controller import Controller, ControllerSettings
from dotbot.logger import setup_logging


@click.command()
@click.option(
    "-p",
    "--port",
    type=str,
    default=SERIAL_PORT_DEFAULT,
    help=f"Virtual com port. Defaults to '{SERIAL_PORT_DEFAULT}'",
)
@click.option(
    "-b",
    "--baudrate",
    type=int,
    default=SERIAL_BAUDRATE_DEFAULT,
    help=f"Serial baudrate. Defaults to {SERIAL_BAUDRATE_DEFAULT}",
)
@click.option(
    "-d",
    "--dotbot-address",
    type=str,
    default=DOTBOT_ADDRESS_DEFAULT,
    help=f"Address in hex of the DotBot to control. Defaults to {DOTBOT_ADDRESS_DEFAULT:>0{16}}",
)
@click.option(
    "-g",
    "--gw-address",
    type=str,
    default=GATEWAY_ADDRESS_DEFAULT,
    help=f"Gateway address in hex. Defaults to {GATEWAY_ADDRESS_DEFAULT:>0{16}}",
)
@click.option(
    "-s",
    "--swarm-id",
    type=str,
    default=SWARM_ID_DEFAULT,
    help=f"Swarm ID in hex. Defaults to {SWARM_ID_DEFAULT:>0{4}}",
)
@click.option(
    "-c",
    "--controller-port",
    type=int,
    default=CONTROLLER_PORT_DEFAULT,
    help=f"Controller port. Defaults to '{CONTROLLER_PORT_DEFAULT}'",
)
@click.option(
    "-w",
    "--webbrowser",
    is_flag=True,
    default=False,
    help="Open a web browser automatically",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Run in verbose mode (all payloads received are printed in terminal)",
)
@click.option(
    "--log-level",
    type=click.Choice(["debug", "info", "warning", "error"]),
    default="info",
    help="Logging level. Defaults to info",
)
@click.option(
    "--log-output",
    type=click.Path(),
    default=os.path.join(os.getcwd(), "pydotbot.log"),
    help="Filename where logs are redirected",
)
@click.option(
    "--handshake",
    is_flag=True,
    default=False,
    help="Perform a basic handshake with the gateway board on startup",
)
@click.option(
    "--edge",
    is_flag=True,
    default=False,
    help="Connect to the edge gateway via MQTT instead of local serial connection",
)
def main(
    port,
    baudrate,
    dotbot_address,
    gw_address,
    swarm_id,
    controller_port,
    webbrowser,
    verbose,
    log_level,
    log_output,
    handshake,
    edge,
):  # pylint: disable=redefined-builtin,too-many-arguments
    """DotBotController, universal SailBot and DotBot controller."""
    # welcome sentence
    print(f"Welcome to the DotBots controller (version: {pydotbot_version()}).")

    setup_logging(log_output, log_level, ["console", "file"])
    try:
        controller = Controller(
            ControllerSettings(
                port=port,
                baudrate=baudrate,
                dotbot_address=dotbot_address,
                gw_address=gw_address,
                swarm_id=swarm_id,
                controller_port=controller_port,
                webbrowser=webbrowser,
                handshake=handshake,
                edge=edge,
                verbose=verbose,
            ),
        )
        asyncio.run(controller.run())
    except serial.serialutil.SerialException as exc:
        sys.exit(f"Serial error: {exc}")
    except (SystemExit, KeyboardInterrupt):
        sys.exit(0)


if __name__ == "__main__":
    main()  # pragma: nocover, pylint: disable=no-value-for-parameter
